"""This module is used to auto-generate server side completion data.

This module can be used during the build process to generate any new
server side autocompletions based on updated service models.
It can also be used to regen completion data as new heuristics are added.

"""

import logging
from collections import defaultdict, namedtuple
from difflib import SequenceMatcher

LOG = logging.getLogger(__name__)
Resource = namedtuple(
    'Resource',
    [
        'resource_name',
        'ident_name',
        'input_parameters',
        'operation',
        'jp_expr',
    ],
)


class ServerCompletionHeuristic:
    # Prefix that typically indicates listing resources,
    # e.g ListResources, DescribeResources, etc.
    _RESOURCE_VERB_PREFIX = ('list', 'describe')
    _OPERATION_EXCLUDES = ('create',)

    def __init__(self, singularize=None):
        if singularize is None:
            singularize = BasicSingularize()
        self._singularize = singularize

    def generate_completion_descriptions(
        self, service_model, prune_completions=True
    ):
        """

        :param service_model: A botocore.model.ServiceModel.
        :param prune_completions: If True, the generated file will be
            pruned of unused resources not used by completion operations.

        """
        # This method is a best-effort attempt at generating server side
        # auto-completion data based on a set of heuristics.  It's not
        # completely accurate and the results should be manually vetted.

        # candidate -> a potential operation that can be used to
        #   to describe a resource (e.g ListResources, DescribeResources)
        # non_candidate -> a potential operation that can use
        #   auto-completion data (e.g DeleteResource, UpdateResource)
        candidates = []
        for op_name in service_model.operation_names:
            if op_name.lower().startswith(self._RESOURCE_VERB_PREFIX):
                candidates.append(op_name)
        all_resources = self._generate_resource_descriptions(
            candidates, service_model
        )
        all_operations = self._generate_operations(
            self._filter_operation_names(service_model.operation_names),
            all_resources,
            service_model,
        )
        if prune_completions:
            self._prune_resource_identifiers(all_resources, all_operations)
        return {
            'version': '1.0',
            'resources': all_resources,
            'operations': all_operations,
        }

    def _filter_operation_names(self, op_names):
        return [
            name
            for name in op_names
            if not name.lower().startswith(self._OPERATION_EXCLUDES)
        ]

    def _generate_resource_descriptions(self, candidates, service_model):
        all_resources = {}
        for op_name in candidates:
            resources = self._resource_for_single_operation(
                op_name, service_model
            )
            if resources is not None:
                for resource in resources:
                    self._inject_resource(all_resources, resource)
        return all_resources

    def _generate_operations(self, op_names, resources, service_model):
        # member_name -> resource
        # This is a simplification we're making for now, it's not at all
        # an actual limitation of auto-generating completion data.
        # If there is a resource that has ``inputParameters`` (i.e required
        # params), then we don't generate completions for you.
        reverse_mapping = defaultdict(list)
        for name, identifier_map in resources.items():
            if identifier_map.get('inputParameters', {}):
                # If there are required inputParameters for a resource
                # we're going to skip them for now.  We should come back
                # to this, it's possible to auto generate these parameters
                # in many cases.
                continue
            for key in identifier_map['resourceIdentifier']:
                reverse_mapping[key].append(name)
        op_map = {}
        for op_name in op_names:
            self._add_completion_data_for_operation(
                op_map, op_name, service_model, reverse_mapping
            )
        return op_map

    def _add_completion_data_for_operation(
        self, op_map, op_name, service_model, reverse_mapping
    ):
        op_model = service_model.operation_model(op_name)
        input_shape = op_model.input_shape
        if not input_shape:
            return
        for member in input_shape.members:
            member_name = self._find_matching_member_name(
                member, reverse_mapping
            )
            if member_name is None:
                continue
            resource_name = self._find_matching_op_name(
                op_name, reverse_mapping[member_name]
            )
            op = op_map.setdefault(op_name, {})
            param = op.setdefault(member, {})
            param['completions'] = [
                {
                    'parameters': {},
                    'resourceName': resource_name,
                    'resourceIdentifier': member_name,
                }
            ]

    def _find_matching_op_name(self, op_name, candidates):
        # If we have more than one operation that gives us back
        # desirable output we match them all with initial operation
        # name and choose the most similar one
        #
        # For example if
        # op_name = DescribeDBInstances and member=DBInstanceIdentifier
        #
        # we can have such list of candidates
        # ['DBInstanceAutomatedBackup', 'DBInstance', 'DBSnapshot']
        #
        # and finally 'DBInstance' will have the highest score

        matcher = SequenceMatcher()
        matcher.set_seq2(op_name)
        matching_score = []
        for candidate in candidates:
            matcher.set_seq1(candidate)
            match_ratio = matcher.ratio()
            matching_score.append((match_ratio, candidate))
        return sorted(matching_score, key=lambda x: x[0], reverse=True)[0][1]

    def _find_matching_member_name(self, member, reverse_mapping):
        # Try to find something in the reverse mapping that's close
        # to the member name.  This is really inefficient but this is
        # a build time step so we don't really care about perf.
        if member in reverse_mapping:
            return member
        matcher = SequenceMatcher()
        matcher.set_seq2(member)
        for candidate in reverse_mapping:
            matcher.set_seq1(candidate)
            match_ratio = matcher.ratio()
            if match_ratio > 0.9:
                return candidate

    def _inject_resource(self, all_resources, resource):
        r = all_resources.setdefault(resource.resource_name, {})
        r['operation'] = resource.operation
        ident = r.setdefault('resourceIdentifier', {})
        if resource.input_parameters:
            r['inputParameters'] = resource.input_parameters
        ident[resource.ident_name] = resource.jp_expr

    def _resource_for_single_operation(self, op_name, service_model):
        # This heuristic works if there's a List/Describe operation
        # that has only one shape in the output that's of type list.
        # It will then find that inspect that list member to find
        # the best guess at which identifier maps to the resource.
        # This is a best effort attempt and relies heavily on naming
        # conventions.
        op_model = service_model.operation_model(op_name)
        output = op_model.output_shape
        list_members = [
            member
            for member, shape in output.members.items()
            if shape.type_name == 'list'
        ]
        if len(list_members) != 1:
            LOG.debug(
                "Operation does not have exactly one list member, "
                "skipping: %s (%s)",
                op_name,
                list_members,
            )
            return
        resource_member_name = list_members[0]
        list_member = output.members[resource_member_name].member
        required_members = []
        if op_model.input_shape is not None:
            required_members = op_model.input_shape.required_members
        if list_member.type_name == 'structure':
            return self._resource_from_structure(
                op_name, resource_member_name, list_member, required_members
            )
        elif list_member.type_name == 'string':
            return [
                self._resource_from_string(
                    op_name,
                    resource_member_name,
                    required_members,
                )
            ]

    def _resource_from_structure(
        self, op_name, resource_member_name, list_member, required_members
    ):
        op_with_prefix_removed = self._remove_verb_prefix(op_name)
        singular_name = self._singularize.make_singular(op_with_prefix_removed)
        resources = []
        for member_name in list_member.members:
            jp_expr = f'{resource_member_name}[].{member_name}'
            r = Resource(
                singular_name, member_name, required_members, op_name, jp_expr
            )
            resources.append(r)
        return resources

    def _resource_from_string(
        self, op_name, resource_member_name, required_members
    ):
        op_with_prefix_removed = self._remove_verb_prefix(op_name)
        singular_name = self._singularize.make_singular(op_with_prefix_removed)
        singular_member_name = self._singularize.make_singular(
            resource_member_name
        )
        r = Resource(
            singular_name,
            singular_member_name,
            required_members,
            op_name,
            f'{resource_member_name}[]',
        )
        return r

    def _remove_verb_prefix(self, op_name):
        for prefix in self._RESOURCE_VERB_PREFIX:
            # 'ListResources' -> 'Resources'
            if op_name.lower().startswith(prefix):
                op_with_prefix_removed = op_name[len(prefix) :]
                return op_with_prefix_removed

    def _prune_resource_identifiers(self, all_resources, all_operations):
        used_identifiers = self._get_identifiers_referenced_by_operations(
            all_operations
        )
        for resource, resource_data in list(all_resources.items()):
            identifiers = resource_data['resourceIdentifier']
            known_ids_for_resource = used_identifiers.get(resource, set())
            for identifier_name in list(identifiers):
                if identifier_name not in known_ids_for_resource:
                    del identifiers[identifier_name]
            if not identifiers:
                # If there's no identifiers used by an autocompletion
                # operation, then we don't need the resource.
                del all_resources[resource]

    def _get_identifiers_referenced_by_operations(self, operations):
        # Dict of resourceName -> Set[resourceIdentifiers]j
        used_identifiers = {}
        for completion in self._all_completions(operations):
            used_identifiers.setdefault(completion['resourceName'], set()).add(
                completion['resourceIdentifier']
            )
        return used_identifiers

    def _all_completions(self, operations):
        for params in operations.values():
            for completions in params.values():
                for completion in completions['completions']:
                    yield completion


class BasicSingularize:
    """Simple implementation of making a word singular.

    Where possible, you should use nltk.stem.WordNetLemmatizer.
    It's not included here as a default to avoid a hard dependency
    on nltk.

    """

    def make_singular(self, name):
        # Certificates -> Certificate
        if name.endswith('ies'):
            return name[:-3] + 'y'
        elif name.endswith('es'):
            return name[:-1]
        elif name.endswith('s'):
            return name[:-1]
        else:
            return name



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Dimension,
    HSplit,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.widgets import Box

from awscli.customizations.wizard.ui.prompt import WizardPrompt


class WizardSectionTab:
    def __init__(self, section_name, section_definition):
        self._name = section_name
        self._definition = section_definition
        self.container = self._get_container()

    def _get_container(self):
        content = f" {self._definition['shortname']}"
        buffer = Buffer(document=Document(content), read_only=True)
        return Window(
            content=BufferControl(buffer=buffer, focusable=False),
            style=self._get_style,
            width=Dimension.exact(len(content) + 1),
            dont_extend_height=True,
        )

    def _get_style(self):
        traverser = get_app().traverser
        if traverser.get_current_section() == self._name:
            return 'class:wizard.section.tab.current'
        elif traverser.has_visited_section(self._name):
            return 'class:wizard.section.tab.visited'
        return 'class:wizard.section.tab.unvisited'

    def __pt_container__(self):
        return self.container


class WizardSectionBody:
    def __init__(self, section_name, section_definition):
        self._name = section_name
        self._definition = section_definition
        self.container = self._get_container()

    def _get_container(self):
        return ConditionalContainer(
            Box(
                HSplit(
                    self._create_prompts_from_section_definition(), padding=1
                ),
                padding_left=2,
                padding_top=1,
            ),
            Condition(self._is_current_section),
        )

    def _is_current_section(self):
        return get_app().traverser.get_current_section() == self._name

    def _create_prompts_from_section_definition(self):
        prompts = []
        for value_name, value_definition in self._definition['values'].items():
            if value_definition['type'] == 'prompt':
                prompts.append(WizardPrompt(value_name, value_definition))
        return prompts

    def __pt_container__(self):
        return self.container



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging

from botocore.exceptions import WaiterError

from awscli.customizations.emr import constants, emrutils, exceptions

LOG = logging.getLogger(__name__)


def validate_and_find_master_dns(session, parsed_globals, cluster_id):
    """
    Utility method for ssh, socks, put and get command.
    Check if the cluster to be connected to is
     terminated or being terminated.
    Check if the cluster is running.
    Find master instance public dns of a given cluster.
    Return the latest created master instance public dns name.
    Throw MasterDNSNotAvailableError or ClusterTerminatedError.
    """
    cluster_state = emrutils.get_cluster_state(
        session, parsed_globals, cluster_id
    )

    if cluster_state in constants.TERMINATED_STATES:
        raise exceptions.ClusterTerminatedError

    emr = emrutils.get_client(session, parsed_globals)

    try:
        cluster_running_waiter = emr.get_waiter('cluster_running')
        if cluster_state in constants.STARTING_STATES:
            print("Waiting for the cluster to start.")
        cluster_running_waiter.wait(ClusterId=cluster_id)
    except WaiterError:
        raise exceptions.MasterDNSNotAvailableError

    return emrutils.find_master_dns(
        session=session, cluster_id=cluster_id, parsed_globals=parsed_globals
    )


def validate_ssh_with_key_file(key_file):
    if (
        emrutils.which('putty.exe')
        or emrutils.which('ssh')
        or emrutils.which('ssh.exe')
    ) is None:
        raise exceptions.SSHNotFoundError
    else:
        check_ssh_key_format(key_file)


def validate_scp_with_key_file(key_file):
    if (
        emrutils.which('pscp.exe')
        or emrutils.which('scp')
        or emrutils.which('scp.exe')
    ) is None:
        raise exceptions.SCPNotFoundError
    else:
        check_scp_key_format(key_file)


def check_scp_key_format(key_file):
    # If only pscp is present and the file format is incorrect
    if (
        emrutils.which('pscp.exe') is not None
        and (emrutils.which('scp.exe') or emrutils.which('scp')) is None
    ):
        if check_command_key_format(key_file, ['ppk']) is False:
            raise exceptions.WrongPuttyKeyError
    else:
        pass


def check_ssh_key_format(key_file):
    # If only putty is present and the file format is incorrect
    if (
        emrutils.which('putty.exe') is not None
        and (emrutils.which('ssh.exe') or emrutils.which('ssh')) is None
    ):
        if check_command_key_format(key_file, ['ppk']) is False:
            raise exceptions.WrongPuttyKeyError
    else:
        pass


def check_command_key_format(key_file, accepted_file_format=[]):
    if any(key_file.endswith(i) for i in accepted_file_format):
        return True
    else:
        return False



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import re
import threading
from io import BytesIO

import awscrt.http
import botocore.awsrequest
import botocore.session
from awscrt.auth import (
    AwsCredentials,
    AwsCredentialsProvider,
    AwsSigningAlgorithm,
    AwsSigningConfig,
)
from awscrt.io import (
    ClientBootstrap,
    ClientTlsContext,
    DefaultHostResolver,
    EventLoopGroup,
    TlsContextOptions,
)
from awscrt.s3 import (
    S3Client,
    S3FileIoOptions,
    S3RequestTlsMode,
    S3RequestType,
    S3ResponseError,
    get_recommended_throughput_target_gbps,
)
from botocore import UNSIGNED
from botocore.compat import urlsplit
from botocore.config import Config
from botocore.exceptions import NoCredentialsError
from botocore.useragent import register_feature_id
from botocore.utils import ArnParser, InvalidArnException, is_s3express_bucket
from s3transfer.constants import FULL_OBJECT_CHECKSUM_ARGS, MB
from s3transfer.exceptions import TransferNotDoneError
from s3transfer.futures import BaseTransferFuture, BaseTransferMeta
from s3transfer.utils import CallArgs, OSUtils, get_callbacks

logger = logging.getLogger(__name__)

CRT_S3_PROCESS_LOCK = None


def acquire_crt_s3_process_lock(name):
    # Currently, the CRT S3 client performs best when there is only one
    # instance of it running on a host. This lock allows an application to
    # signal across processes whether there is another process of the same
    # application using the CRT S3 client and prevent spawning more than one
    # CRT S3 clients running on the system for that application.
    #
    # NOTE: When acquiring the CRT process lock, the lock automatically is
    # released when the lock object is garbage collected. So, the CRT process
    # lock is set as a global so that it is not unintentionally garbage
    # collected/released if reference of the lock is lost.
    global CRT_S3_PROCESS_LOCK
    if CRT_S3_PROCESS_LOCK is None:
        crt_lock = awscrt.s3.CrossProcessLock(name)
        try:
            crt_lock.acquire()
        except RuntimeError:
            # If there is another process that is holding the lock, the CRT
            # returns a RuntimeError. We return None here to signal that our
            # current process was not able to acquire the lock.
            return None
        CRT_S3_PROCESS_LOCK = crt_lock
    return CRT_S3_PROCESS_LOCK


def create_s3_crt_client(
    region,
    crt_credentials_provider=None,
    num_threads=None,
    target_throughput=None,
    part_size=8 * MB,
    use_ssl=True,
    verify=None,
    fio_options=None,
):
    """
    :type region: str
    :param region: The region used for signing

    :type crt_credentials_provider:
        Optional[awscrt.auth.AwsCredentialsProvider]
    :param crt_credentials_provider: CRT AWS credentials provider
        to use to sign requests. If not set, requests will not be signed.

    :type num_threads: Optional[int]
    :param num_threads: Number of worker threads generated. Default
        is the number of processors in the machine.

    :type target_throughput: Optional[int]
    :param target_throughput: Throughput target in bytes per second.
        By default, CRT will automatically attempt to choose a target
        throughput that matches the system's maximum network throughput.
        Currently, if CRT is unable to determine the maximum network
        throughput, a fallback target throughput of ``1_250_000_000`` bytes
        per second (which translates to 10 gigabits per second, or 1.16
        gibibytes per second) is used. To set a specific target
        throughput, set a value for this parameter.

    :type part_size: Optional[int]
    :param part_size: Size, in Bytes, of parts that files will be downloaded
        or uploaded in.

    :type use_ssl: boolean
    :param use_ssl: Whether or not to use SSL.  By default, SSL is used.
        Note that not all services support non-ssl connections.

    :type verify: Optional[boolean/string]
    :param verify: Whether or not to verify SSL certificates.
        By default SSL certificates are verified.  You can provide the
        following values:

        * False - do not validate SSL certificates.  SSL will still be
            used (unless use_ssl is False), but SSL certificates
            will not be verified.
        * path/to/cert/bundle.pem - A filename of the CA cert bundle to
            use. Specify this argument if you want to use a custom CA cert
            bundle instead of the default one on your system.

    :type fio_options: Optional[dict]
    :param fio_options: Kwargs to use to build an `awscrt.s3.S3FileIoOptions`.
    """

    event_loop_group = EventLoopGroup(num_threads)
    host_resolver = DefaultHostResolver(event_loop_group)
    bootstrap = ClientBootstrap(event_loop_group, host_resolver)
    tls_connection_options = None

    tls_mode = (
        S3RequestTlsMode.ENABLED if use_ssl else S3RequestTlsMode.DISABLED
    )
    if verify is not None:
        tls_ctx_options = TlsContextOptions()
        if verify:
            tls_ctx_options.override_default_trust_store_from_path(
                ca_filepath=verify
            )
        else:
            tls_ctx_options.verify_peer = False
        client_tls_option = ClientTlsContext(tls_ctx_options)
        tls_connection_options = client_tls_option.new_connection_options()
    target_gbps = _get_crt_throughput_target_gbps(
        provided_throughput_target_bytes=target_throughput
    )
    crt_fio_options = None
    if fio_options:
        crt_fio_options = S3FileIoOptions(**fio_options)
    return S3Client(
        bootstrap=bootstrap,
        region=region,
        credential_provider=crt_credentials_provider,
        part_size=part_size,
        tls_mode=tls_mode,
        tls_connection_options=tls_connection_options,
        throughput_target_gbps=target_gbps,
        enable_s3express=True,
        fio_options=crt_fio_options,
    )


def _get_crt_throughput_target_gbps(provided_throughput_target_bytes=None):
    if provided_throughput_target_bytes is None:
        target_gbps = get_recommended_throughput_target_gbps()
        logger.debug(
            'Recommended CRT throughput target in gbps: %s', target_gbps
        )
        if target_gbps is None:
            target_gbps = 10.0
    else:
        # NOTE: The GB constant in s3transfer is technically a gibibyte. The
        # GB constant is not used here because the CRT interprets gigabits
        # for networking as a base power of 10
        # (i.e. 1000 ** 3 instead of 1024 ** 3).
        target_gbps = provided_throughput_target_bytes * 8 / 1_000_000_000
    logger.debug('Using CRT throughput target in gbps: %s', target_gbps)
    return target_gbps


class CRTTransferManager:
    def __init__(self, crt_s3_client, crt_request_serializer, osutil=None):
        """A transfer manager interface for Amazon S3 on CRT s3 client.

        :type crt_s3_client: awscrt.s3.S3Client
        :param crt_s3_client: The CRT s3 client, handling all the
            HTTP requests and functions under then hood

        :type crt_request_serializer: s3transfer.crt.BaseCRTRequestSerializer
        :param crt_request_serializer: Serializer, generates unsigned crt HTTP
            request.

        :type osutil: s3transfer.utils.OSUtils
        :param osutil: OSUtils object to use for os-related behavior when
            using with transfer manager.
        """
        if osutil is None:
            self._osutil = OSUtils()
        self._crt_s3_client = crt_s3_client
        self._s3_args_creator = S3ClientArgsCreator(
            crt_request_serializer, self._osutil
        )
        self._crt_exception_translator = (
            crt_request_serializer.translate_crt_exception
        )
        self._future_coordinators = []
        self._semaphore = threading.Semaphore(128)  # not configurable
        # A counter to create unique id's for each transfer submitted.
        self._id_counter = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, *args):
        cancel = False
        if exc_type:
            cancel = True
        self._shutdown(cancel)

    def download(
        self, bucket, key, fileobj, extra_args=None, subscribers=None
    ):
        if extra_args is None:
            extra_args = {}
        if subscribers is None:
            subscribers = {}
        callargs = CallArgs(
            bucket=bucket,
            key=key,
            fileobj=fileobj,
            extra_args=extra_args,
            subscribers=subscribers,
        )
        return self._submit_transfer("get_object", callargs)

    def upload(self, fileobj, bucket, key, extra_args=None, subscribers=None):
        if extra_args is None:
            extra_args = {}
        if subscribers is None:
            subscribers = {}
        self._validate_checksum_algorithm_supported(extra_args)
        callargs = CallArgs(
            bucket=bucket,
            key=key,
            fileobj=fileobj,
            extra_args=extra_args,
            subscribers=subscribers,
        )
        return self._submit_transfer("put_object", callargs)

    def delete(self, bucket, key, extra_args=None, subscribers=None):
        if extra_args is None:
            extra_args = {}
        if subscribers is None:
            subscribers = {}
        callargs = CallArgs(
            bucket=bucket,
            key=key,
            extra_args=extra_args,
            subscribers=subscribers,
        )
        return self._submit_transfer("delete_object", callargs)

    def shutdown(self, cancel=False):
        self._shutdown(cancel)

    def _validate_checksum_algorithm_supported(self, extra_args):
        checksum_algorithm = extra_args.get('ChecksumAlgorithm')
        if checksum_algorithm is None:
            return
        supported_algorithms = list(awscrt.s3.S3ChecksumAlgorithm.__members__)
        if checksum_algorithm.upper() not in supported_algorithms:
            raise ValueError(
                f'ChecksumAlgorithm: {checksum_algorithm} not supported. '
                f'Supported algorithms are: {supported_algorithms}'
            )

    def _cancel_transfers(self):
        for coordinator in self._future_coordinators:
            if not coordinator.done():
                coordinator.cancel()

    def _finish_transfers(self):
        for coordinator in self._future_coordinators:
            coordinator.result()

    def _wait_transfers_done(self):
        for coordinator in self._future_coordinators:
            coordinator.wait_until_on_done_callbacks_complete()

    def _shutdown(self, cancel=False):
        if cancel:
            self._cancel_transfers()
        try:
            self._finish_transfers()

        except KeyboardInterrupt:
            self._cancel_transfers()
        except Exception:
            pass
        finally:
            self._wait_transfers_done()

    def _release_semaphore(self, **kwargs):
        self._semaphore.release()

    def _submit_transfer(self, request_type, call_args):
        register_feature_id('S3_TRANSFER')
        on_done_after_calls = [self._release_semaphore]
        coordinator = CRTTransferCoordinator(
            transfer_id=self._id_counter,
            exception_translator=self._crt_exception_translator,
        )
        components = {
            'meta': CRTTransferMeta(self._id_counter, call_args),
            'coordinator': coordinator,
        }
        future = CRTTransferFuture(**components)
        afterdone = AfterDoneHandler(coordinator)
        on_done_after_calls.append(afterdone)

        try:
            self._semaphore.acquire()
            on_queued = self._s3_args_creator.get_crt_callback(
                future, 'queued'
            )
            on_queued()
            crt_callargs = self._s3_args_creator.get_make_request_args(
                request_type,
                call_args,
                coordinator,
                future,
                on_done_after_calls,
            )
            crt_s3_request = self._crt_s3_client.make_request(**crt_callargs)
        except Exception as e:
            coordinator.set_exception(e, True)
            on_done = self._s3_args_creator.get_crt_callback(
                future, 'done', after_subscribers=on_done_after_calls
            )
            on_done(error=e)
        else:
            coordinator.set_s3_request(crt_s3_request)
        self._future_coordinators.append(coordinator)

        self._id_counter += 1
        return future


class CRTTransferMeta(BaseTransferMeta):
    """Holds metadata about the CRTTransferFuture"""

    def __init__(self, transfer_id=None, call_args=None):
        self._transfer_id = transfer_id
        self._call_args = call_args
        self._user_context = {}

    @property
    def call_args(self):
        return self._call_args

    @property
    def transfer_id(self):
        return self._transfer_id

    @property
    def user_context(self):
        return self._user_context


class CRTTransferFuture(BaseTransferFuture):
    def __init__(self, meta=None, coordinator=None):
        """The future associated to a submitted transfer request via CRT S3 client

        :type meta: s3transfer.crt.CRTTransferMeta
        :param meta: The metadata associated to the transfer future.

        :type coordinator: s3transfer.crt.CRTTransferCoordinator
        :param coordinator: The coordinator associated to the transfer future.
        """
        self._meta = meta
        if meta is None:
            self._meta = CRTTransferMeta()
        self._coordinator = coordinator

    @property
    def meta(self):
        return self._meta

    def done(self):
        return self._coordinator.done()

    def result(self, timeout=None):
        self._coordinator.result(timeout)

    def cancel(self):
        self._coordinator.cancel()

    def set_exception(self, exception):
        """Sets the exception on the future."""
        if not self.done():
            raise TransferNotDoneError(
                'set_exception can only be called once the transfer is '
                'complete.'
            )
        self._coordinator.set_exception(exception, override=True)


class BaseCRTRequestSerializer:
    def serialize_http_request(self, transfer_type, future):
        """Serialize CRT HTTP requests.

        :type transfer_type: string
        :param transfer_type: the type of transfer made,
            e.g 'put_object', 'get_object', 'delete_object'

        :type future: s3transfer.crt.CRTTransferFuture

        :rtype: awscrt.http.HttpRequest
        :returns: An unsigned HTTP request to be used for the CRT S3 client
        """
        raise NotImplementedError('serialize_http_request()')

    def translate_crt_exception(self, exception):
        raise NotImplementedError('translate_crt_exception()')


class BotocoreCRTRequestSerializer(BaseCRTRequestSerializer):
    def __init__(self, session, client_kwargs=None):
        """Serialize CRT HTTP request using botocore logic
        It also takes into account configuration from both the session
        and any keyword arguments that could be passed to
        `Session.create_client()` when serializing the request.

        :type session: botocore.session.Session

        :type client_kwargs: Optional[Dict[str, str]])
        :param client_kwargs: The kwargs for the botocore
            s3 client initialization.
        """
        self._session = session
        if client_kwargs is None:
            client_kwargs = {}
        self._resolve_client_config(session, client_kwargs)
        self._client = session.create_client(**client_kwargs)
        self._client.meta.events.register(
            'request-created.s3.*', self._capture_http_request
        )
        self._client.meta.events.register(
            'after-call.s3.*', self._change_response_to_serialized_http_request
        )
        self._client.meta.events.register(
            'before-send.s3.*', self._make_fake_http_response
        )
        self._client.meta.events.register(
            'before-call.s3.*', self._remove_checksum_context
        )

    def _resolve_client_config(self, session, client_kwargs):
        user_provided_config = None
        if session.get_default_client_config():
            user_provided_config = session.get_default_client_config()
        if 'config' in client_kwargs:
            user_provided_config = client_kwargs['config']

        client_config = Config(signature_version=UNSIGNED)
        if user_provided_config:
            client_config = user_provided_config.merge(client_config)
        client_kwargs['config'] = client_config
        client_kwargs["service_name"] = "s3"

    def _crt_request_from_aws_request(self, aws_request):
        url_parts = urlsplit(aws_request.url)
        crt_path = url_parts.path
        if url_parts.query:
            crt_path = f'{crt_path}?{url_parts.query}'
        headers_list = []
        for name, value in aws_request.headers.items():
            if isinstance(value, str):
                headers_list.append((name, value))
            else:
                headers_list.append((name, str(value, 'utf-8')))

        crt_headers = awscrt.http.HttpHeaders(headers_list)

        crt_request = awscrt.http.HttpRequest(
            method=aws_request.method,
            path=crt_path,
            headers=crt_headers,
            body_stream=aws_request.body,
        )
        return crt_request

    def _convert_to_crt_http_request(self, botocore_http_request):
        # Logic that does CRTUtils.crt_request_from_aws_request
        crt_request = self._crt_request_from_aws_request(botocore_http_request)
        if crt_request.headers.get("host") is None:
            # If host is not set, set it for the request before using CRT s3
            url_parts = urlsplit(botocore_http_request.url)
            crt_request.headers.set("host", url_parts.netloc)
        if crt_request.headers.get('Content-MD5') is not None:
            crt_request.headers.remove("Content-MD5")

        # In general, the CRT S3 client expects a content length header. It
        # only expects a missing content length header if the body is not
        # seekable. However, botocore does not set the content length header
        # for GetObject API requests and so we set the content length to zero
        # to meet the CRT S3 client's expectation that the content length
        # header is set even if there is no body.
        if crt_request.headers.get('Content-Length') is None:
            if botocore_http_request.body is None:
                crt_request.headers.add('Content-Length', "0")

        # Botocore sets the Transfer-Encoding header when it cannot determine
        # the content length of the request body (e.g. it's not seekable).
        # However, CRT does not support this header, but it supports
        # non-seekable bodies. So we remove this header to not cause issues
        # in the downstream CRT S3 request.
        if crt_request.headers.get('Transfer-Encoding') is not None:
            crt_request.headers.remove('Transfer-Encoding')

        return crt_request

    def _capture_http_request(self, request, **kwargs):
        request.context['http_request'] = request

    def _change_response_to_serialized_http_request(
        self, context, parsed, **kwargs
    ):
        request = context['http_request']
        parsed['HTTPRequest'] = request.prepare()

    def _make_fake_http_response(self, request, **kwargs):
        return botocore.awsrequest.AWSResponse(
            None,
            200,
            {},
            FakeRawResponse(b""),
        )

    def _get_botocore_http_request(self, client_method, call_args):
        return getattr(self._client, client_method)(
            Bucket=call_args.bucket, Key=call_args.key, **call_args.extra_args
        )['HTTPRequest']

    def serialize_http_request(self, transfer_type, future):
        botocore_http_request = self._get_botocore_http_request(
            transfer_type, future.meta.call_args
        )
        crt_request = self._convert_to_crt_http_request(botocore_http_request)
        return crt_request

    def translate_crt_exception(self, exception):
        if isinstance(exception, S3ResponseError):
            return self._translate_crt_s3_response_error(exception)
        else:
            return None

    def _translate_crt_s3_response_error(self, s3_response_error):
        status_code = s3_response_error.status_code
        if status_code < 301:
            # Botocore's exception parsing only
            # runs on status codes >= 301
            return None

        headers = {k: v for k, v in s3_response_error.headers}
        operation_name = s3_response_error.operation_name
        if operation_name is not None:
            service_model = self._client.meta.service_model
            shape = service_model.operation_model(operation_name).output_shape
        else:
            shape = None

        response_dict = {
            'headers': botocore.awsrequest.HeadersDict(headers),
            'status_code': status_code,
            'body': s3_response_error.body,
        }
        parsed_response = self._client._response_parser.parse(
            response_dict, shape=shape
        )

        error_code = parsed_response.get("Error", {}).get("Code")
        error_class = self._client.exceptions.from_code(error_code)
        return error_class(parsed_response, operation_name=operation_name)

    def _remove_checksum_context(self, params, **kwargs):
        request_context = params.get("context", {})
        if "checksum" in request_context:
            del request_context["checksum"]


class FakeRawResponse(BytesIO):
    def stream(self, amt=1024, decode_content=None):
        while True:
            chunk = self.read(amt)
            if not chunk:
                break
            yield chunk


class BotocoreCRTCredentialsWrapper:
    def __init__(self, resolved_botocore_credentials):
        self._resolved_credentials = resolved_botocore_credentials

    def __call__(self):
        credentials = self._get_credentials().get_frozen_credentials()
        return AwsCredentials(
            credentials.access_key, credentials.secret_key, credentials.token
        )

    def to_crt_credentials_provider(self):
        return AwsCredentialsProvider.new_delegate(self)

    def _get_credentials(self):
        if self._resolved_credentials is None:
            raise NoCredentialsError()
        return self._resolved_credentials


class CRTTransferCoordinator:
    """A helper class for managing CRTTransferFuture"""

    def __init__(
        self, transfer_id=None, s3_request=None, exception_translator=None
    ):
        self.transfer_id = transfer_id
        self._exception_translator = exception_translator
        self._s3_request = s3_request
        self._lock = threading.Lock()
        self._exception = None
        self._crt_future = None
        self._done_event = threading.Event()

    @property
    def s3_request(self):
        return self._s3_request

    def set_done_callbacks_complete(self):
        self._done_event.set()

    def wait_until_on_done_callbacks_complete(self, timeout=None):
        self._done_event.wait(timeout)

    def set_exception(self, exception, override=False):
        with self._lock:
            if not self.done() or override:
                self._exception = exception

    def cancel(self):
        if self._s3_request:
            self._s3_request.cancel()

    def result(self, timeout=None):
        if self._exception:
            raise self._exception
        try:
            self._crt_future.result(timeout)
        except KeyboardInterrupt:
            self.cancel()
            self._crt_future.result(timeout)
            raise
        except Exception as e:
            self.handle_exception(e)
        finally:
            if self._s3_request:
                self._s3_request = None

    def handle_exception(self, exc):
        translated_exc = None
        if self._exception_translator:
            try:
                translated_exc = self._exception_translator(exc)
            except Exception as e:
                # Bail out if we hit an issue translating
                # and raise the original error.
                logger.debug("Unable to translate exception.", exc_info=e)
                pass
        if translated_exc is not None:
            raise translated_exc from exc
        else:
            raise exc

    def done(self):
        if self._crt_future is None:
            return False
        return self._crt_future.done()

    def set_s3_request(self, s3_request):
        self._s3_request = s3_request
        self._crt_future = self._s3_request.finished_future


class S3ClientArgsCreator:
    def __init__(self, crt_request_serializer, os_utils):
        self._request_serializer = crt_request_serializer
        self._os_utils = os_utils

    def get_make_request_args(
        self, request_type, call_args, coordinator, future, on_done_after_calls
    ):
        request_args_handler = getattr(
            self,
            f'_get_make_request_args_{request_type}',
            self._default_get_make_request_args,
        )
        return request_args_handler(
            request_type=request_type,
            call_args=call_args,
            coordinator=coordinator,
            future=future,
            on_done_before_calls=[],
            on_done_after_calls=on_done_after_calls,
        )

    def get_crt_callback(
        self,
        future,
        callback_type,
        before_subscribers=None,
        after_subscribers=None,
    ):
        def invoke_all_callbacks(*args, **kwargs):
            callbacks_list = []
            if before_subscribers is not None:
                callbacks_list += before_subscribers
            callbacks_list += get_callbacks(future, callback_type)
            if after_subscribers is not None:
                callbacks_list += after_subscribers
            for callback in callbacks_list:
                # The get_callbacks helper will set the first augment
                # by keyword, the other augments need to be set by keyword
                # as well
                if callback_type == "progress":
                    callback(bytes_transferred=args[0])
                else:
                    callback(*args, **kwargs)

        return invoke_all_callbacks

    def _get_make_request_args_put_object(
        self,
        request_type,
        call_args,
        coordinator,
        future,
        on_done_before_calls,
        on_done_after_calls,
    ):
        send_filepath = None
        if isinstance(call_args.fileobj, str):
            send_filepath = call_args.fileobj
            data_len = self._os_utils.get_file_size(send_filepath)
            call_args.extra_args["ContentLength"] = data_len
        else:
            call_args.extra_args["Body"] = call_args.fileobj

        checksum_config = None
        if not any(
            checksum_arg in call_args.extra_args
            for checksum_arg in FULL_OBJECT_CHECKSUM_ARGS
        ):
            checksum_algorithm = call_args.extra_args.pop(
                'ChecksumAlgorithm', 'CRC64NVME'
            ).upper()
            checksum_config = awscrt.s3.S3ChecksumConfig(
                algorithm=awscrt.s3.S3ChecksumAlgorithm[checksum_algorithm],
                location=awscrt.s3.S3ChecksumLocation.TRAILER,
            )
        # Suppress botocore's automatic MD5 calculation by setting an override
        # value that will get deleted in the BotocoreCRTRequestSerializer.
        # As part of the CRT S3 request, we request the CRT S3 client to
        # automatically add trailing checksums to its uploads.
        call_args.extra_args["ContentMD5"] = "override-to-be-removed"

        make_request_args = self._default_get_make_request_args(
            request_type=request_type,
            call_args=call_args,
            coordinator=coordinator,
            future=future,
            on_done_before_calls=on_done_before_calls,
            on_done_after_calls=on_done_after_calls,
        )
        make_request_args['send_filepath'] = send_filepath
        make_request_args['checksum_config'] = checksum_config
        return make_request_args

    def _get_make_request_args_get_object(
        self,
        request_type,
        call_args,
        coordinator,
        future,
        on_done_before_calls,
        on_done_after_calls,
    ):
        recv_filepath = None
        on_body = None
        checksum_config = awscrt.s3.S3ChecksumConfig(validate_response=True)
        if isinstance(call_args.fileobj, str):
            final_filepath = call_args.fileobj
            recv_filepath = self._os_utils.get_temp_filename(final_filepath)
            on_done_before_calls.append(
                RenameTempFileHandler(
                    coordinator, final_filepath, recv_filepath, self._os_utils
                )
            )
        else:
            on_body = OnBodyFileObjWriter(call_args.fileobj)

        make_request_args = self._default_get_make_request_args(
            request_type=request_type,
            call_args=call_args,
            coordinator=coordinator,
            future=future,
            on_done_before_calls=on_done_before_calls,
            on_done_after_calls=on_done_after_calls,
        )
        make_request_args['recv_filepath'] = recv_filepath
        make_request_args['on_body'] = on_body
        make_request_args['checksum_config'] = checksum_config
        return make_request_args

    def _default_get_make_request_args(
        self,
        request_type,
        call_args,
        coordinator,
        future,
        on_done_before_calls,
        on_done_after_calls,
    ):
        make_request_args = {
            'request': self._request_serializer.serialize_http_request(
                request_type, future
            ),
            'type': getattr(
                S3RequestType, request_type.upper(), S3RequestType.DEFAULT
            ),
            'on_done': self.get_crt_callback(
                future, 'done', on_done_before_calls, on_done_after_calls
            ),
            'on_progress': self.get_crt_callback(future, 'progress'),
        }

        # For DEFAULT requests, CRT requires the official S3 operation name.
        # So transform string like "delete_object" -> "DeleteObject".
        if make_request_args['type'] == S3RequestType.DEFAULT:
            make_request_args['operation_name'] = ''.join(
                x.title() for x in request_type.split('_')
            )

        arn_handler = _S3ArnParamHandler()
        if (
            accesspoint_arn_details := arn_handler.handle_arn(call_args.bucket)
        ) and accesspoint_arn_details['region'] == "":
            # Configure our region to `*` to propogate in `x-amz-region-set`
            # for multi-region support in MRAP accesspoints.
            # use_double_uri_encode and should_normalize_uri_path are defaulted to be True
            # But SDK already encoded the URI, and it's for S3, so set both to False
            make_request_args['signing_config'] = AwsSigningConfig(
                algorithm=AwsSigningAlgorithm.V4_ASYMMETRIC,
                region="*",
                use_double_uri_encode=False,
                should_normalize_uri_path=False,
            )
            call_args.bucket = accesspoint_arn_details['resource_name']
        elif is_s3express_bucket(call_args.bucket):
            # use_double_uri_encode and should_normalize_uri_path are defaulted to be True
            # But SDK already encoded the URI, and it's for S3, so set both to False
            make_request_args['signing_config'] = AwsSigningConfig(
                algorithm=AwsSigningAlgorithm.V4_S3EXPRESS,
                use_double_uri_encode=False,
                should_normalize_uri_path=False,
            )
        return make_request_args


class RenameTempFileHandler:
    def __init__(self, coordinator, final_filename, temp_filename, osutil):
        self._coordinator = coordinator
        self._final_filename = final_filename
        self._temp_filename = temp_filename
        self._osutil = osutil

    def __call__(self, **kwargs):
        error = kwargs['error']
        if error:
            self._osutil.remove_file(self._temp_filename)
        else:
            try:
                self._osutil.rename_file(
                    self._temp_filename, self._final_filename
                )
            except Exception as e:
                self._osutil.remove_file(self._temp_filename)
                # the CRT future has done already at this point
                self._coordinator.set_exception(e)


class AfterDoneHandler:
    def __init__(self, coordinator):
        self._coordinator = coordinator

    def __call__(self, **kwargs):
        self._coordinator.set_done_callbacks_complete()


class OnBodyFileObjWriter:
    def __init__(self, fileobj):
        self._fileobj = fileobj

    def __call__(self, chunk, **kwargs):
        self._fileobj.write(chunk)


class _S3ArnParamHandler:
    """Partial port of S3ArnParamHandler from botocore.

    This is used to make a determination on MRAP accesspoints for signing
    purposes. This should be safe to remove once we properly integrate auth
    resolution from Botocore into the CRT transfer integration.
    """

    _RESOURCE_REGEX = re.compile(
        r'^(?P<resource_type>accesspoint|outpost)[/:](?P<resource_name>.+)$'
    )

    def __init__(self):
        self._arn_parser = ArnParser()

    def handle_arn(self, bucket):
        arn_details = self._get_arn_details_from_bucket(bucket)
        if arn_details is None:
            return
        if arn_details['resource_type'] == 'accesspoint':
            return arn_details

    def _get_arn_details_from_bucket(self, bucket):
        try:
            arn_details = self._arn_parser.parse_arn(bucket)
            self._add_resource_type_and_name(arn_details)
            return arn_details
        except InvalidArnException:
            pass
        return None

    def _add_resource_type_and_name(self, arn_details):
        match = self._RESOURCE_REGEX.match(arn_details['resource'])
        if match:
            arn_details['resource_type'] = match.group('resource_type')
            arn_details['resource_name'] = match.group('resource_name')



# Copyright 2012-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os


def make_url(region, bucket_name, obj_path, version=None):
    """
    This link describes the format of Path Style URLs
    http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingBucket.html#access-bucket-intro
    """
    base = "https://s3.amazonaws.com"
    if region and region != "us-east-1":
        base = f"https://s3-{region}.amazonaws.com"

    result = f"{base}/{bucket_name}/{obj_path}"
    if version:
        result = f"{result}?versionId={version}"

    return result


def get_s3_path(file_path):
    return os.path.basename(file_path)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from botocore import model

from awscli.arguments import BaseCLIArgument

logger = logging.getLogger(__name__)


DEFAULT = 1
HELP = (
    """
<p>Number of instances to launch. If a single number is provided, it
is assumed to be the minimum to launch (defaults to %d).  If a range is
provided in the form <code>min:max</code> then the first number is
interpreted as the minimum number of instances to launch and the second
is interpreted as the maximum number of instances to launch.</p>"""
    % DEFAULT
)


def register_count_events(event_handler):
    event_handler.register(
        'building-argument-table.ec2.run-instances', ec2_add_count
    )
    event_handler.register(
        'before-parameter-build.ec2.RunInstances', set_default_count
    )


def ec2_add_count(argument_table, **kwargs):
    argument_table['count'] = CountArgument('count')
    del argument_table['min-count']
    del argument_table['max-count']


def set_default_count(params, **kwargs):
    params.setdefault('MaxCount', DEFAULT)
    params.setdefault('MinCount', DEFAULT)


class CountArgument(BaseCLIArgument):
    def __init__(self, name):
        self.argument_model = model.Shape('CountArgument', {'type': 'string'})
        self._name = name
        self._required = False

    @property
    def cli_name(self):
        return '--' + self._name

    @property
    def cli_type_name(self):
        return 'string'

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return HELP

    def add_to_parser(self, parser):
        # We do NOT set default value here. It will be set later by event hook.
        parser.add_argument(
            self.cli_name,
            metavar=self.py_name,
            help='Number of instances to launch',
        )

    def add_to_params(self, parameters, value):
        if value is None:
            # NO-OP if value is not explicitly set by user
            return
        try:
            if ':' in value:
                minstr, maxstr = value.split(':')
            else:
                minstr, maxstr = (value, value)
            parameters['MinCount'] = int(minstr)
            parameters['MaxCount'] = int(maxstr)
        except:
            msg = (
                'count parameter should be of '
                'form min[:max] (e.g. 1 or 1:10)'
            )
            raise ParamValidationError(msg)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys

from awscli.customizations.ecs.expressgateway.color_utils import ColorUtils
from awscli.customizations.ecs.expressgateway.managedresource import (
    ManagedResource,
)

MANAGED_RESOURCE_EMPTY = "{indent}<empty>"


class ManagedResourceGroup:
    """
    Represents a logical grouping of ManagedResources.
    """

    RESOURCE_GROUP_HEADER = "{resource_type} - {identifier}"
    RESOURCE_LIST_HEADER = "{resource_type}"

    def __init__(
        self,
        resource_type=None,
        identifier=None,
        resources=[],
        status=None,
        reason=None,
    ):
        self.resource_type = resource_type
        self.identifier = identifier
        # Maintain input ordering
        self.sorted_resource_keys = [
            self._create_key(resource) for resource in resources
        ]
        self.resource_mapping = {
            self._create_key(resource): resource for resource in resources
        }
        self.status = status
        self.reason = reason
        self.color_utils = ColorUtils()

    def _create_key(self, resource):
        resource_type = (
            resource.resource_type if resource.resource_type else ""
        )
        identifier = resource.identifier if resource.identifier else ""
        return resource_type + "/" + identifier

    def get_stream_string(self, timestamp, use_color=True):
        """Returns flattened stream strings for all resources in the group.

        Args:
            timestamp (str): Timestamp string to prefix each resource
            use_color (bool): Whether to use ANSI color codes (default: True)

        Returns:
            str: All flattened resources formatted for stream display, separated by newlines
        """
        flat_resources = []

        for resource in self.resource_mapping.values():
            if isinstance(resource, ManagedResourceGroup):
                # Recursively flatten nested groups
                nested = resource.get_stream_string(timestamp, use_color)
                if nested:
                    flat_resources.append(nested)
            elif isinstance(resource, ManagedResource):
                # Get stream string for individual resource
                flat_resources.append(
                    resource.get_stream_string(timestamp, use_color)
                )

        return "\n".join(flat_resources)

    def get_changed_resources(self, previous_resources_dict):
        """Get flattened list of resources that have changed properties.

        Compares individual resource properties (status, reason, updated_at, etc.)
        against previous state to detect changes. This is used for change detection
        in TEXT-ONLY mode, NOT for DEPLOYMENT diff (use compare_resource_sets for that).

        Args:
            previous_resources_dict: Dict of {(resource_type, identifier): ManagedResource}
                from previous poll. Can be empty dict for first poll.

        Returns:
            tuple: (changed_resources, updated_dict, removed_keys)
                - changed_resources: List of ManagedResource that changed or None if no changes
                - updated_dict: Updated dict with current resources for next comparison
                - removed_keys: Set of keys that were removed since last poll
        """
        current_resources = self._flatten_to_list()
        changed_resources = []
        updated_dict = {}

        for resource in current_resources:
            resource_key = (resource.resource_type, resource.identifier)
            previous_resource = previous_resources_dict.get(resource_key)

            if not previous_resource:
                changed_resources.append(resource)
            else:
                if resource.compare_properties(previous_resource):
                    changed_resources.append(resource)

            updated_dict[resource_key] = resource

        current_keys = {
            (r.resource_type, r.identifier) for r in current_resources
        }
        removed_keys = set(previous_resources_dict.keys()) - current_keys

        return (
            changed_resources if changed_resources else None,
            updated_dict,
            removed_keys,
        )

    def _flatten_to_list(self):
        """Flatten this resource group into a list of individual resources."""
        flat_list = []
        for resource in self.resource_mapping.values():
            if isinstance(resource, ManagedResourceGroup):
                flat_list.extend(resource._flatten_to_list())
            elif isinstance(resource, ManagedResource):
                flat_list.append(resource)
        return flat_list

    def is_terminal(self):
        return not self.resource_mapping or all(
            [
                resource.is_terminal()
                for resource in self.resource_mapping.values()
            ]
        )

    def get_status_string(self, spinner_char, depth=0, use_color=True):
        """Returns the resource information as a string.

        Args:
            spinner_char (str): Character to display for in-progress resources
            depth (int): Indentation depth for nested display (default: 0)
            use_color (bool): Whether to use ANSI color codes (default: True)

        Returns:
            str: Formatted status string with resource information
        """
        lines = []

        if self.resource_type:
            header = " " * depth + self.color_utils.make_cyan(
                str(self.resource_type), use_color
            )

            if self.identifier:
                header += ": "
                if self.status:
                    header += self.color_utils.make_status_symbol(
                        self.status, spinner_char, use_color
                    )
                    header += self.color_utils.color_by_status(
                        self.identifier, self.status, use_color
                    )
                    header += " - " + self.color_utils.color_by_status(
                        self.status, self.status, use_color
                    )
                else:
                    header += self.identifier
            elif self.status:
                header += " " + self.color_utils.make_status_symbol(
                    self.status, spinner_char, use_color
                )
                header += "- " + self.color_utils.color_by_status(
                    self.status, self.status, use_color
                )

            lines.append(header)

            if self.status and self.reason:
                lines.append(" " * (depth + 1) + "Reason: " + self.reason)

            if self.sorted_resource_keys:
                # Add a new line to help with spacing
                lines.append("")

        depth_offset = 1 if self.resource_type else 0
        for key in self.sorted_resource_keys:
            resource_string = self.resource_mapping.get(key).get_status_string(
                spinner_char, depth=depth + depth_offset, use_color=use_color
            )
            lines.append(resource_string)

        if not self.sorted_resource_keys and self.resource_type:
            empty_line = MANAGED_RESOURCE_EMPTY.format(
                indent=" " * (depth + depth_offset)
            )
            lines.append(empty_line)

        return '\n'.join(lines)

    def combine(self, other_resource_group):
        """Returns a ManagedResourceGroup containing the combined responses of each resource.

        Args:
            other_resource_group (ManagedResourceGroup): Resource group to combine with this one

        Returns:
            ManagedResourceGroup: Combined resource group with merged resources
        """
        if not other_resource_group:
            return self

        # Collect all resources from both groups
        all_resources = list(self.resource_mapping.values()) + list(
            other_resource_group.resource_mapping.values()
        )

        # Group by resource type, prioritizing resources with identifiers
        type_groups = {}
        for resource in all_resources:
            resource_type = resource.resource_type
            if resource_type not in type_groups:
                type_groups[resource_type] = []
            type_groups[resource_type].append(resource)

        # For each type, if we have both resources with and without identifiers, keep only those with identifiers
        filtered_resources = []
        for resource_type, resources in type_groups.items():
            resources_with_id = [r for r in resources if r.identifier]
            resources_without_id = [r for r in resources if not r.identifier]

            if resources_with_id and resources_without_id:
                # Only keep resources with identifiers
                filtered_resources.extend(resources_with_id)
            else:
                # Keep all resources (either all have identifiers or all don't)
                filtered_resources.extend(resources)

        # Now create keys and combine as before
        resource_mapping = {}
        for resource in filtered_resources:
            key = self._create_key(resource)
            if key in resource_mapping:
                resource_mapping[key] = resource_mapping[key].combine(resource)
            else:
                resource_mapping[key] = resource

        return ManagedResourceGroup(
            resource_type=self.resource_type,
            identifier=self.identifier,
            resources=resource_mapping.values(),
        )

    def _combine_child_resources(self, resource_a, resource_b):
        if resource_a:
            return resource_a.combine(resource_b)
        else:
            return resource_b

    def compare_resource_sets(self, other_resource_group):
        """Compares resource SETS between two groups to find additions/removals.

        This is used for DEPLOYMENT view to show which resources were added or removed
        between service configurations, NOT for detecting property changes within
        individual resources (that's compare_properties() in ManagedResource).

        Args:
            other_resource_group (ManagedResourceGroup): Resource group to compare against

        Returns:
            tuple: (self_unique, other_unique) where:
                - self_unique (ManagedResourceGroup): Resources unique to this group
                - other_unique (ManagedResourceGroup): Resources unique to the other group
        """
        other_keys = set(other_resource_group.resource_mapping.keys())
        self_keys = set(self.resource_mapping.keys())

        unique_to_self = self_keys - other_keys

        # Get resource types from self that have no identifier (end with '/')
        self_types_without_id = {
            key.rstrip('/') for key in self_keys if key.endswith('/')
        }

        # Exclude keys from other if their type matches a type in self without identifier
        unique_to_other = {
            key
            for key in (other_keys - self_keys)
            if not any(key.startswith(t + '/') for t in self_types_without_id)
        }
        common_keys = self_keys & other_keys

        common_diff = {
            key: self.resource_mapping[key].compare_resource_sets(
                other_resource_group.resource_mapping.get(key)
            )
            for key in common_keys
            if not isinstance(self.resource_mapping[key], ManagedResource)
        }
        common_self = {
            key: val[0]
            for key, val in common_diff.items()
            if val[0] is not None
        }
        common_other = {
            key: val[1]
            for key, val in common_diff.items()
            if val[1] is not None
        }

        self_resources = [
            self.resource_mapping[key]
            if key in unique_to_self
            else common_self[key]
            for key in self.sorted_resource_keys
            if key in unique_to_self or key in common_self
        ]

        other_resources = [
            other_resource_group.resource_mapping[key]
            if key in unique_to_other
            else common_other[key]
            for key in other_resource_group.sorted_resource_keys
            if key in unique_to_other or key in common_other
        ]

        return (
            ManagedResourceGroup(
                resource_type=self.resource_type,
                identifier=self.identifier,
                resources=self_resources,
            ),
            ManagedResourceGroup(
                resource_type=self.resource_type,
                identifier=self.identifier,
                resources=other_resources,
            ),
        )



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import sys

from botocore.hooks import HierarchicalEmitter

log = logging.getLogger('awscli.plugin')

BUILTIN_PLUGINS = {'__builtin__': 'awscli.handlers'}
CLI_LEGACY_PLUGIN_PATH = 'cli_legacy_plugin_path'


def load_plugins(plugin_mapping, event_hooks=None, include_builtins=True):
    """

    :type plugin_mapping: dict
    :param plugin_mapping: A dict of plugin name to import path,
        e.g. ``{"plugingName": "package.modulefoo"}``.

    :type event_hooks: ``EventHooks``
    :param event_hooks: Event hook emitter.  If one if not provided,
        an emitter will be created and returned.  Otherwise, the
        passed in ``event_hooks`` will be used to initialize plugins.

    :type include_builtins: bool
    :param include_builtins: If True, the builtin awscli plugins (specified in
        ``BUILTIN_PLUGINS``) will be included in the list of plugins to load.

    :rtype: HierarchicalEmitter
    :return: An event emitter object.

    """
    if event_hooks is None:
        event_hooks = HierarchicalEmitter()
    if include_builtins:
        _load_plugins(BUILTIN_PLUGINS, event_hooks)
    plugin_path = plugin_mapping.pop(CLI_LEGACY_PLUGIN_PATH, None)
    if plugin_path is not None:
        _add_plugin_path_to_sys_path(plugin_path)
        _load_plugins(plugin_mapping, event_hooks)
    else:
        log.debug(
            "cli_legacy_plugin_path not defined in plugin section. Not "
            "importing additional plugins."
        )
    return event_hooks


def _load_plugins(plugin_mapping, event_hooks):
    modules = _import_plugins(plugin_mapping)
    for name, plugin in zip(plugin_mapping.keys(), modules):
        log.debug("Initializing plugin %s: %s", name, plugin)
        plugin.awscli_initialize(event_hooks)


def _import_plugins(plugin_mapping):
    plugins = []
    for name, path in plugin_mapping.items():
        log.debug("Importing plugin %s: %s", name, path)
        if '.' not in path:
            plugins.append(__import__(path))
        else:
            package, module = path.rsplit('.', 1)
            module = __import__(path, fromlist=[module])
            plugins.append(module)
    return plugins


def _add_plugin_path_to_sys_path(plugin_path):
    for dirname in plugin_path.split(os.pathsep):
        log.debug(
            "Adding additional path from cli_legacy_plugin_path "
            "configuration: %s",
            dirname,
        )
        sys.path.append(dirname)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import base64
import binascii
import datetime
import hashlib
import json
import logging
import secrets
import string
import textwrap
import uuid
from enum import Enum

from awscrt.crypto import EC, ECExportFormat, ECType
from botocore.compat import parse_qsl
from botocore.useragent import register_feature_id
from dateutil.tz.tz import tzutc

from awscli.botocore.exceptions import (
    LoginAuthorizationCodeError,
    LoginInsufficientPermissions,
)
from awscli.botocore.utils import (
    ArnParser,
    build_add_dpop_header_handler,
    get_base_sign_in_uri,
    percent_encode_sequence,
)
from awscli.compat import compat_input

LOG = logging.getLogger(__name__)


class LoginType(Enum):
    SAME_DEVICE = 1
    CROSS_DEVICE = 2


CLIENT_ID = {
    LoginType.SAME_DEVICE: 'arn:aws:signin:::devtools/same-device',
    LoginType.CROSS_DEVICE: 'arn:aws:signin:::devtools/cross-device',
}


class BaseLoginTokenFetcher:
    """Base class that can acquire an access token from AWS Sign-In"""

    def __init__(
        self,
        client,
        on_pending_authorization,
        private_key=None,
        base_uri_builder=None,
    ):
        self._client = client
        self._on_pending_authorization = on_pending_authorization
        self._expected_state = uuid.uuid4()

        self._code_verifier = ''.join(
            secrets.choice(string.ascii_letters + string.digits + '-._~')
            for _ in range(64)
        )
        self._code_challenge = base64.urlsafe_b64encode(
            hashlib.sha256(self._code_verifier.encode()).digest()
        ).decode()

        if private_key is None:
            private_key = EC.new_generate(ECType.P_256)
        self._private_key = private_key

        if base_uri_builder is None:
            base_uri_builder = get_base_sign_in_uri
        self._base_endpoint = base_uri_builder(self._client)

    def fetch_token(self):
        """Retrieves an access token via AWS Sign-In"""
        raise NotImplementedError('fetch_token')

    @staticmethod
    def _get_browser_handler_args(authorization_uri):
        """Generates the arguments for the method that opens the browser"""
        return {
            'verificationUri': authorization_uri,
            'verificationUriComplete': authorization_uri,
            'userCode': None,
            'cross_device_flag': '--remote',
        }

    def _exchange_auth_code_for_access_token(
        self, client_id, auth_code, redirect_uri=None
    ):
        """Exchanges the temporary auth code for an access token"""
        body = {
            'clientId': client_id,
            'grantType': 'authorization_code',
            'code': auth_code,
            'codeVerifier': self._code_verifier,
        }

        # Won't have a redirect URI for the cross-device workflow
        if redirect_uri:
            body['redirectUri'] = redirect_uri

        self._client.meta.events.register(
            'before-call.signin.CreateOAuth2Token',
            build_add_dpop_header_handler(self._private_key),
        )

        try:
            response = self._client.create_o_auth2_token(
                tokenInput=body,
            )
        except self._client.exceptions.AccessDeniedException as e:
            error_type = e.response.get('error', '')
            if error_type == 'AUTHCODE_EXPIRED':
                raise LoginAuthorizationCodeError(
                    error_msg="Unable to complete the login process due to an expired authorization "
                    "code. Please reauthenticate using 'aws login'."
                )
            elif error_type == 'INSUFFICIENT_PERMISSIONS':
                raise LoginInsufficientPermissions()

            raise

        if response is None or 'tokenOutput' not in response:
            raise LoginAuthorizationCodeError(
                error_msg='Failed to retrieve an access token.'
            )

        output = response.get('tokenOutput')

        expires_timestamp = datetime.datetime.now(
            tzutc()
        ) + datetime.timedelta(seconds=output['expiresIn'])
        login_session = self._extract_login_session_from_id_token(
            output['idToken']
        )
        account_id = self._extract_account_id_from_login_session(login_session)

        token = {
            'accessToken': {
                'accessKeyId': output['accessToken']['accessKeyId'],
                'secretAccessKey': output['accessToken']['secretAccessKey'],
                'sessionToken': output['accessToken']['sessionToken'],
                'accountId': account_id,
                'expiresAt': expires_timestamp.strftime('%Y-%m-%dT%H:%M:%SZ'),
            },
            'tokenType': output['tokenType'],
            'clientId': client_id,
            'refreshToken': output['refreshToken'],
            'idToken': output['idToken'],
            'dpopKey': self._serialize_to_pem(self._private_key),
        }
        return token, login_session

    def _get_authorization_uri(
        self, client_id, expected_state, code_challenge, redirect_uri=None
    ):
        query_params = {
            'response_type': 'code',
            'client_id': client_id,
            'state': expected_state,
            'code_challenge_method': 'SHA-256',
            'scope': 'openid',
            # Don't want to encode code_challenge again, so we append below
        }

        if redirect_uri:
            query_params['redirect_uri'] = redirect_uri

        return (
            f'{self._base_endpoint}/v1/authorize?'
            f'{percent_encode_sequence(query_params)}'
            f'&code_challenge={code_challenge[:-1]}'
        )

    @staticmethod
    def _extract_login_session_from_id_token(id_token):
        # Extract the subject arn from the JWT payload
        parts = id_token.split('.')
        if len(parts) != 3:
            raise ValueError('Invalid JWT token - cannot extract payload')
        payload_bytes = base64_url_decode(parts[1])
        payload_json = json.loads(payload_bytes)

        if 'sub' not in payload_json:
            raise ValueError('Invalid JWT token - missing sub')

        return payload_json['sub']

    @staticmethod
    def _extract_account_id_from_login_session(login_session):
        login_session_arn = ArnParser().parse_arn(login_session)

        return login_session_arn['account']

    @staticmethod
    def _serialize_to_pem(private_key):
        contents = base64.b64encode(
            private_key.export_key(ECExportFormat.SEC1)
        ).decode('utf-8')
        # RFC 1421 calls for 64 character lines max
        wrapped_contents = textwrap.fill(contents, width=64)
        return f'-----BEGIN EC PRIVATE KEY-----\n{wrapped_contents}\n-----END EC PRIVATE KEY-----\n'


class SameDeviceLoginTokenFetcher(BaseLoginTokenFetcher):
    """
    Logs in using OAuth's auth code and PKCE flow,
    intended for when the user is on a device with a web browser and the
    ability to initialize the callback server
    """

    def __init__(
        self,
        client,
        auth_code_fetcher,
        on_pending_authorization,
        private_key=None,
        base_uri_builder=None,
    ):
        super().__init__(
            client,
            on_pending_authorization,
            private_key=private_key,
            base_uri_builder=base_uri_builder,
        )
        self._auth_code_fetcher = auth_code_fetcher

    def fetch_token(self):
        register_feature_id('LOGIN_SAME_DEVICE')
        authorization_uri = self._get_authorization_uri(
            client_id=CLIENT_ID[LoginType.SAME_DEVICE],
            redirect_uri=self._auth_code_fetcher.redirect_uri_with_port(),
            expected_state=self._expected_state,
            code_challenge=self._code_challenge,
        )

        # Open/display the link, then block until the redirect uri is hit
        self._on_pending_authorization(
            **self._get_browser_handler_args(authorization_uri)
        )
        auth_code, state = self._auth_code_fetcher.get_auth_code_and_state()

        if auth_code is None:
            raise LoginAuthorizationCodeError(
                error_msg='Failed to retrieve an authorization code.'
            )

        # The state we get back from the redirect is just a string, so
        # cast our original UUID before comparing
        if state != str(self._expected_state):
            raise LoginAuthorizationCodeError(
                error_msg=f'State parameter {state} does not match expected value {self._expected_state}.'
            )

        return self._exchange_auth_code_for_access_token(
            client_id=CLIENT_ID[LoginType.SAME_DEVICE],
            auth_code=auth_code,
            redirect_uri=self._auth_code_fetcher.redirect_uri_with_port(),
        )


class CrossDeviceLoginTokenFetcher(BaseLoginTokenFetcher):
    """
    Logs in by prompting the user to enter the auth code,
    intended for when the user is unable to open a web browser
    on the same device where the CLI is running.
    """

    def fetch_token(self):
        register_feature_id('LOGIN_CROSS_DEVICE')
        redirect_uri = f'{self._base_endpoint}/v1/sessions/confirmation'

        authorization_uri = self._get_authorization_uri(
            client_id=CLIENT_ID[LoginType.CROSS_DEVICE],
            expected_state=self._expected_state,
            code_challenge=self._code_challenge,
            redirect_uri=redirect_uri,
        )

        self._on_pending_authorization(
            **self._get_browser_handler_args(authorization_uri)
        )

        verification_code = self._prompt(
            '\nEnter the authorization code displayed in your browser'
        )

        auth_code, state = self.parse_verification_code(verification_code)

        if auth_code is None:
            raise LoginAuthorizationCodeError(
                error_msg='Failed to retrieve an authorization code.'
            )

        if state != str(self._expected_state):
            raise LoginAuthorizationCodeError(
                error_msg=f'State parameter {state} does not match expected value {self._expected_state}.'
            )

        return self._exchange_auth_code_for_access_token(
            client_id=CLIENT_ID[LoginType.CROSS_DEVICE],
            auth_code=auth_code,
            redirect_uri=redirect_uri,
        )

    @staticmethod
    def parse_verification_code(verification_code):
        """
        Parse the verification code that the user pastes from the browser,
        which is expected to be base64-encoded 'state={state}&auth_code={code}'
        """
        try:
            query_string = base64.b64decode(verification_code).decode('utf-8')
        except (UnicodeDecodeError, binascii.Error):
            raise ValueError('Failed to decode the verification code.')

        params_dict = dict(parse_qsl(query_string))

        if 'state' not in params_dict or 'code' not in params_dict:
            raise ValueError(
                'Failed to retrieve an auth_code or state from the verification code'
            )

        return params_dict['code'], params_dict['state']

    @staticmethod
    def _prompt(
        prompt_text,
    ):
        response = compat_input(f'{prompt_text}: ')
        if not response:
            response = None
        return response


def base64_url_decode(data):
    data += '=' * (-len(data) % 4)  # restore '=' padding
    return base64.urlsafe_b64decode(data)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.application import get_app
from prompt_toolkit.filters import Condition


@Condition
def doc_section_visible():
    app = get_app()
    return getattr(app, 'show_doc')


@Condition
def help_section_visible():
    app = get_app()
    return getattr(app, 'show_help')


@Condition
def output_section_visible():
    app = get_app()
    return app.show_output


@Condition
def is_multi_column():
    app = get_app()
    return getattr(app, 'multi_column', False)


@Condition
def is_one_column():
    app = get_app()
    return not getattr(app, 'multi_column', False)


@Condition
def doc_window_has_focus():
    "Only activate these key bindings if doc window has focus."
    app = get_app()
    return app.current_buffer.name == 'doc_buffer'


@Condition
def search_input_has_focus():
    "Only activate these key bindings if doc window search has focus."
    app = get_app()
    return app.layout.current_window.style == 'class:search-toolbar'


@Condition
def input_buffer_has_focus():
    "Only activate these key bindings if input buffer has focus."
    app = get_app()
    return app.current_buffer.name == 'input_buffer'


@Condition
def is_history_mode():
    """Only activate these key bindings if input buffer has focus
    and history_mode is on"""
    buffer = get_app().current_buffer
    return buffer.name == 'input_buffer' and buffer.history_mode


@Condition
def is_debug_mode():
    """Only activate these key bindings if input buffer has focus
    and history_mode is on"""
    app = get_app()
    return app.debug



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.model import Shape

from awscli.arguments import BaseCLIArgument


def add_streaming_output_arg(
    argument_table, operation_model, session, **kwargs
):
    # Implementation detail:  hooked up to 'building-argument-table'
    # event.
    if _has_streaming_output(operation_model):
        streaming_argument_name = _get_streaming_argument_name(operation_model)
        argument_table['outfile'] = StreamingOutputArgument(
            response_key=streaming_argument_name,
            operation_model=operation_model,
            session=session,
            name='outfile',
        )


def _has_streaming_output(model):
    return model.has_streaming_output


def _get_streaming_argument_name(model):
    return model.output_shape.serialization['payload']


class StreamingOutputArgument(BaseCLIArgument):
    BUFFER_SIZE = 32768
    HELP = 'Filename where the content will be saved'

    def __init__(
        self, response_key, operation_model, name, session, buffer_size=None
    ):
        self._name = name
        self.argument_model = Shape(
            'StreamingOutputArgument', {'type': 'string'}
        )
        if buffer_size is None:
            buffer_size = self.BUFFER_SIZE
        self._buffer_size = buffer_size
        # This is the key in the response body where we can find the
        # streamed contents.
        self._response_key = response_key
        self._output_file = None
        self._name = name
        self._required = True
        self._operation_model = operation_model
        self._session = session

    @property
    def cli_name(self):
        # Because this is a parameter, not an option, it shouldn't have the
        # '--' prefix. We want to use the self.py_name to indicate that it's an
        # argument.
        return self._name

    @property
    def cli_type_name(self):
        return 'string'

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return self.HELP

    def add_to_parser(self, parser):
        parser.add_argument(self._name, metavar=self.py_name, help=self.HELP)

    def add_to_params(self, parameters, value):
        self._output_file = value
        service_id = self._operation_model.service_model.service_id.hyphenize()
        operation_name = self._operation_model.name
        self._session.register(
            'after-call.%s.%s' % (service_id, operation_name), self.save_file
        )

    def save_file(self, parsed, **kwargs):
        if self._response_key not in parsed:
            # If the response key is not in parsed, then
            # we've received an error message and we'll let the AWS CLI
            # error handler print out an error message.  We have no
            # file to save in this situation.
            return
        body = parsed[self._response_key]
        buffer_size = self._buffer_size
        with open(self._output_file, 'wb') as fp:
            data = body.read(buffer_size)
            while data:
                fp.write(data)
                data = body.read(buffer_size)
        # We don't want to include the streaming param in
        # the returned response.
        del parsed[self._response_key]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.commands import BasicCommand


def register_dev_commands(event_handlers):
    event_handlers.register(
        'building-command-table.main', CLIDevCommand.add_command
    )


# This is adding a top level placeholder command to add dev commands.
# Each specific customization can add to this command via the
# building-command-table event name.
class CLIDevCommand(BasicCommand):
    NAME = 'cli-dev'
    DESCRIPTION = (
        'Internal commands for AWS CLI development.\n'
        'These commands are not intended for normal end usage. These '
        'commands are used to help develop and debug the AWS CLI.\n'
        'Do not rely on these commands, backwards compatibility '
        'is not guaranteed.  Any of these commands under "aws cli-dev" '
        'may be removed in future versions.\n'
    )
    SYNOPSIS = 'aws cli-dev [dev-command]'



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import json
import sys
import xml.dom.minidom
import xml.parsers.expat

import colorama

from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.history.commands import HistorySubcommand
from awscli.customizations.history.filters import RegexFilter
from awscli.table import COLORAMA_KWARGS


class Formatter:
    def __init__(self, output=None, include=None, exclude=None):
        """Formats and outputs CLI history events

        :type output: File-like obj
        :param output: The stream to write the formatted event to. By default
            sys.stdout is used.

        :type include: list
        :param include: A filter specifying which event to only be displayed.
            This parameter is mutually exclusive with exclude.

        :type exclude: list
        :param exclude: A filter specifying which events to exclude from being
            displayed. This parameter is mutually exclusive with include.

        """
        self._output = output
        if self._output is None:
            self._output = sys.stdout
        if include and exclude:
            raise ParamValidationError(
                'Either input or exclude can be provided but not both'
            )
        self._include = include
        self._exclude = exclude

    def display(self, event_record):
        """Displays a formatted version of the event record

        :type event_record: dict
        :param event_record: The event record to format and display.
        """
        if self._should_display(event_record):
            self._display(event_record)

    def _display(self, event_record):
        raise NotImplementedError('_display()')

    def _should_display(self, event_record):
        if self._include:
            return event_record['event_type'] in self._include
        elif self._exclude:
            return event_record['event_type'] not in self._exclude
        else:
            return True


class DetailedFormatter(Formatter):
    _SIG_FILTER = RegexFilter(
        'Signature=([a-z0-9]{4})[a-z0-9]{60}',
        r'Signature=\1...',
    )

    _SECTIONS = {
        'CLI_VERSION': {
            'title': 'AWS CLI command entered',
            'values': [{'description': 'with AWS CLI version'}],
        },
        'CLI_ARGUMENTS': {'values': [{'description': 'with arguments'}]},
        'API_CALL': {
            'title': 'API call made',
            'values': [
                {'description': 'to service', 'payload_key': 'service'},
                {'description': 'using operation', 'payload_key': 'operation'},
                {
                    'description': 'with parameters',
                    'payload_key': 'params',
                    'value_format': 'dictionary',
                },
            ],
        },
        'HTTP_REQUEST': {
            'title': 'HTTP request sent',
            'values': [
                {'description': 'to URL', 'payload_key': 'url'},
                {'description': 'with method', 'payload_key': 'method'},
                {
                    'description': 'with headers',
                    'payload_key': 'headers',
                    'value_format': 'dictionary',
                    'filters': [_SIG_FILTER],
                },
                {
                    'description': 'with body',
                    'payload_key': 'body',
                    'value_format': 'http_body',
                },
            ],
        },
        'HTTP_RESPONSE': {
            'title': 'HTTP response received',
            'values': [
                {
                    'description': 'with status code',
                    'payload_key': 'status_code',
                },
                {
                    'description': 'with headers',
                    'payload_key': 'headers',
                    'value_format': 'dictionary',
                },
                {
                    'description': 'with body',
                    'payload_key': 'body',
                    'value_format': 'http_body',
                },
            ],
        },
        'PARSED_RESPONSE': {
            'title': 'HTTP response parsed',
            'values': [
                {'description': 'parsed to', 'value_format': 'dictionary'}
            ],
        },
        'CLI_RC': {
            'title': 'AWS CLI command exited',
            'values': [{'description': 'with return code'}],
        },
    }

    _COMPONENT_COLORS = {
        'title': colorama.Style.BRIGHT,
        'description': colorama.Fore.CYAN,
    }

    def __init__(self, output=None, include=None, exclude=None, colorize=True):
        super(DetailedFormatter, self).__init__(output, include, exclude)
        self._request_id_to_api_num = {}
        self._num_api_calls = 0
        self._colorize = colorize
        self._value_pformatter = SectionValuePrettyFormatter()
        if self._colorize:
            colorama.init(**COLORAMA_KWARGS)  # noqa

    def _display(self, event_record):
        section_definition = self._SECTIONS.get(event_record['event_type'])
        if section_definition is not None:
            self._display_section(event_record, section_definition)

    def _display_section(self, event_record, section_definition):
        if 'title' in section_definition:
            self._display_title(section_definition['title'], event_record)
        for value_definition in section_definition['values']:
            self._display_value(value_definition, event_record)

    def _display_title(self, title, event_record):
        formatted_title = self._format_section_title(title, event_record)
        self._write_output(formatted_title)

    def _display_value(self, value_definition, event_record):
        value_description = value_definition['description']
        event_record_payload = event_record['payload']
        value = event_record_payload
        if 'payload_key' in value_definition:
            value = event_record_payload[value_definition['payload_key']]
        formatted_value = self._format_description(value_description)
        formatted_value += self._format_value(
            value, event_record, value_definition.get('value_format')
        )
        if 'filters' in value_definition:
            for text_filter in value_definition['filters']:
                formatted_value = text_filter.filter_text(formatted_value)
        self._write_output(formatted_value)

    def _write_output(self, content):
        self._output.write(content)

    def _format_section_title(self, title, event_record):
        formatted_title = title
        api_num = self._get_api_num(event_record)
        if api_num is not None:
            formatted_title = ('[%s] ' % api_num) + formatted_title
        formatted_title = self._color_if_configured(formatted_title, 'title')
        formatted_title += '\n'

        formatted_timestamp = self._format_description('at time')
        formatted_timestamp += self._format_value(
            event_record['timestamp'], event_record, value_format='timestamp'
        )

        return '\n' + formatted_title + formatted_timestamp

    def _get_api_num(self, event_record):
        request_id = event_record['request_id']
        if request_id:
            if request_id not in self._request_id_to_api_num:
                self._request_id_to_api_num[request_id] = self._num_api_calls
                self._num_api_calls += 1
            return self._request_id_to_api_num[request_id]

    def _format_description(self, value_description):
        return self._color_if_configured(
            value_description + ': ', 'description'
        )

    def _format_value(self, value, event_record, value_format=None):
        if value_format:
            formatted_value = self._value_pformatter.pformat(
                value, value_format, event_record
            )
        else:
            formatted_value = str(value)
        return formatted_value + '\n'

    def _color_if_configured(self, text, component):
        if self._colorize:
            color = self._COMPONENT_COLORS[component]
            return color + text + colorama.Style.RESET_ALL
        return text


class SectionValuePrettyFormatter:
    def pformat(self, value, value_format, event_record):
        return getattr(self, '_pformat_' + value_format)(value, event_record)

    def _pformat_timestamp(self, event_timestamp, event_record=None):
        return datetime.datetime.fromtimestamp(
            event_timestamp / 1000.0
        ).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

    def _pformat_dictionary(self, obj, event_record=None):
        return json.dumps(obj=obj, sort_keys=True, indent=4)

    def _pformat_http_body(self, body, event_record):
        if not body:
            return 'There is no associated body'
        elif event_record['payload'].get('streaming', False):
            return 'The body is a stream and will not be displayed'
        elif self._is_xml(body):
            # TODO: Figure out a way to minimize the number of times we have
            # to parse the XML. Currently at worst, it will take three times.
            # One to determine if it is XML, another to strip whitespace, and
            # a third to convert to make it pretty. This is an issue as it
            # can cause issues when there are large XML payloads such as
            # an s3 ListObjects call.
            return self._get_pretty_xml(body)
        elif self._is_json_structure(body):
            return self._get_pretty_json(body)
        else:
            return body

    def _get_pretty_xml(self, body):
        # The body is parsed and whitespace is stripped because some services
        # like ec2 already return pretty XML and if toprettyxml() was applied
        # to it, it will add even more newlines and spaces on top of it.
        # So this just removes all whitespace from the start to prevent the
        # chance of adding to much newlines and spaces when toprettyxml()
        # is called.
        stripped_body = self._strip_whitespace(body)
        xml_dom = xml.dom.minidom.parseString(stripped_body)
        return xml_dom.toprettyxml(indent=' ' * 4, newl='\n')

    def _get_pretty_json(self, body):
        # The json body is loaded so it can be dumped in a format that
        # is desired.
        obj = json.loads(body)
        return self._pformat_dictionary(obj)

    def _is_xml(self, body):
        try:
            xml.dom.minidom.parseString(body)
        except xml.parsers.expat.ExpatError:
            return False
        return True

    def _strip_whitespace(self, xml_string):
        xml_dom = xml.dom.minidom.parseString(xml_string)
        return ''.join([line.strip() for line in xml_dom.toxml().splitlines()])

    def _is_json_structure(self, body):
        if body.startswith('{'):
            try:
                json.loads(body)
                return True
            except json.decoder.JSONDecodeError:
                return False
        return False


class ShowCommand(HistorySubcommand):
    NAME = 'show'
    DESCRIPTION = (
        'Shows the various events related to running a specific CLI command. '
        'If this command is ran without any positional arguments, it will '
        'display the events for the last CLI command ran.'
    )
    FORMATTERS = {'detailed': DetailedFormatter}
    ARG_TABLE = [
        {
            'name': 'command_id',
            'nargs': '?',
            'default': 'latest',
            'positional_arg': True,
            'help_text': (
                'The ID of the CLI command to show. If this positional argument '
                'is omitted, it will show the last the CLI command ran.'
            ),
        },
        {
            'name': 'include',
            'nargs': '+',
            'help_text': (
                'Specifies which events to **only** include when showing the '
                'CLI command. This argument is mutually exclusive with '
                '``--exclude``.'
            ),
        },
        {
            'name': 'exclude',
            'nargs': '+',
            'help_text': (
                'Specifies which events to exclude when showing the '
                'CLI command. This argument is mutually exclusive with '
                '``--include``.'
            ),
        },
        {
            'name': 'format',
            'choices': FORMATTERS.keys(),
            'default': 'detailed',
            'help_text': (
                'Specifies which format to use in showing the events for '
                'the specified CLI command. The following formats are '
                'supported:\n\n'
                '<ul>'
                '<li> detailed - This the default format. It prints out a '
                'detailed overview of the CLI command ran. It displays all '
                'of the key events in the command lifecycle where each '
                'important event has a title and its important values '
                'underneath. The events are ordered by timestamp and events of '
                'the same API call are associated together with the '
                '[``api_id``] notation where events that share the same '
                '``api_id`` belong to the lifecycle of the same API call.'
                '</li>'
                '</ul>'
            ),
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        self._connect_to_history_db()
        try:
            self._validate_args(parsed_args)
            with self._output_stream_factory.get_output_stream() as stream:
                formatter = self._get_formatter(
                    parsed_args, parsed_globals, stream
                )
                for record in self._get_record_iterator(parsed_args):
                    formatter.display(record)
        finally:
            self._close_history_db()
        return 0

    def _validate_args(self, parsed_args):
        if parsed_args.exclude and parsed_args.include:
            raise ParamValidationError(
                'Either --exclude or --include can be provided but not both'
            )

    def _get_formatter(self, parsed_args, parsed_globals, output_stream):
        format_type = parsed_args.format
        formatter_kwargs = {
            'include': parsed_args.include,
            'exclude': parsed_args.exclude,
            'output': output_stream,
        }
        if format_type == 'detailed':
            formatter_kwargs['colorize'] = self._should_use_color(
                parsed_globals
            )
        return self.FORMATTERS[format_type](**formatter_kwargs)

    def _get_record_iterator(self, parsed_args):
        if parsed_args.command_id == 'latest':
            return self._db_reader.iter_latest_records()
        else:
            return self._db_reader.iter_records(parsed_args.command_id)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""
AWS CLI ECS Express Gateway Service Monitoring Module.

This module provides real-time monitoring capabilities for Amazon ECS Express Gateway Services,
allowing users to track resource creation progress, deployment status, and service health through
an interactive command-line interface with live updates and visual indicators.

The data collection logic is handled by ServiceViewCollector, which parses AWS resources and
formats monitoring output. This module focuses on display and user interaction.

The module implements two resource view modes:
- RESOURCE: Displays all resources associated with the service
- DEPLOYMENT: Shows resources that have changed in the most recent deployment

And two display modes:
- INTERACTIVE: Real-time display with spinner and keyboard navigation (requires TTY)
- TEXT-ONLY: Text output with timestamps and change detection (works without TTY)

Key Features:
- Real-time progress monitoring with spinner animations
- Diff-based resource tracking for deployment changes
- Comprehensive resource parsing (load balancers, security groups, auto-scaling, etc.)
- Graceful keyboard interrupt handling

Classes:
    ECSMonitorExpressGatewayService: Main CLI command class for service monitoring
    ECSExpressGatewayServiceWatcher: Core monitoring logic and resource tracking

Usage:
    aws ecs monitor-express-gateway-service --service-arn <arn> [--resource-view RESOURCE|DEPLOYMENT] [--mode INTERACTIVE|TEXT-ONLY]
"""

import sys
import time

from botocore.exceptions import ClientError

from awscli.customizations.commands import BasicCommand
from awscli.customizations.ecs.exceptions import MonitoringError
from awscli.customizations.ecs.expressgateway.display_strategy import (
    InteractiveDisplayStrategy,
    TextOnlyDisplayStrategy,
)
from awscli.customizations.ecs.prompt_toolkit_display import Display
from awscli.customizations.ecs.serviceviewcollector import ServiceViewCollector
from awscli.customizations.utils import uni_print


class ECSMonitorExpressGatewayService(BasicCommand):
    """AWS CLI command for monitoring ECS Express Gateway Service deployments.

    Provides real-time monitoring of service resource creation and deployment
    progress with interactive output and status updates.
    """

    NAME = 'monitor-express-gateway-service'

    DESCRIPTION = (
        "Monitors the progress of resource creation for an ECS Express Gateway Service. "
        "This command provides real-time monitoring of service deployments showing the status "
        "of load balancers, security groups, auto-scaling "
        "configurations, and other AWS resources as they are created or updated. "
        "Use ``--resource-view RESOURCE`` to view all service resources, or ``--resource-view DEPLOYMENT`` to track only "
        "resources that have changed in the most recent deployment. "
        "Choose ``--mode INTERACTIVE`` for real-time display with keyboard navigation (requires TTY), "
        "or ``--mode TEXT-ONLY`` for text output with timestamps (works without TTY). "
        "The monitoring session continues until manually stopped by the user or the specified timeout is reached. "
        "In INTERACTIVE mode, use keyboard shortcuts: up/down to scroll through resources, 'q' to quit. "
        "In TEXT-ONLY mode, press Ctrl+C to stop monitoring."
    )

    ARG_TABLE = [
        {
            'name': 'service-arn',
            'help_text': (
                "The short name or full Amazon Resource Name "
                "(ARN) of the service to monitor."
            ),
            'required': True,
        },
        {
            'name': 'resource-view',
            'help_text': (
                "Specifies which resources to display during monitoring. "
                "RESOURCE (default) - Combines resources from all active configurations of the service "
                "and displays the live status of load balancers, security groups, target groups, etc. "
                "DEPLOYMENT - Determines the resources that are being added or removed as part of the "
                "latest service deployment, and displays their live statuses."
            ),
            'required': False,
            'default': 'RESOURCE',
            'choices': ['RESOURCE', 'DEPLOYMENT'],
        },
        {
            'name': 'mode',
            'help_text': (
                "Display mode for monitoring output. "
                "INTERACTIVE (default if TTY available) - Real-time display with spinner and keyboard navigation. "
                "TEXT-ONLY - Text output with timestamps and change detection (works without TTY)."
            ),
            'required': False,
            'choices': ['INTERACTIVE', 'TEXT-ONLY'],
        },
        {
            'name': 'timeout',
            'help_text': (
                "Maximum time in minutes to monitor the service. "
                "Default is 30 minutes."
            ),
            'required': False,
            'default': 30,
            'cli_type_name': 'integer',
        },
    ]

    def __init__(self, session, watcher_class=None):
        """Initialize the command with optional dependency injection.

        Args:
            session: AWS session
            watcher_class: Optional watcher class for dependency injection (for testing)
        """
        super().__init__(session)
        self._watcher_class = watcher_class or ECSExpressGatewayServiceWatcher

    def _run_main(self, parsed_args, parsed_globals):
        """Execute the monitoring command with parsed arguments.

        Args:
            parsed_args: Command line arguments including service-arn, resource_view, and timeout
            parsed_globals: Global CLI configuration including region and endpoint
        """
        try:
            display_mode = self._determine_display_mode(parsed_args.mode)
        except ValueError as e:
            uni_print(f"aws: [ERROR]: {str(e)}", sys.stderr)
            return 1

        try:
            self._client = self._session.create_client(
                'ecs',
                region_name=parsed_globals.region,
                endpoint_url=parsed_globals.endpoint_url,
                verify=parsed_globals.verify_ssl,
            )

            # Determine if color should be used
            use_color = self._should_use_color(parsed_globals)

            self._watcher_class(
                self._client,
                parsed_args.service_arn,
                parsed_args.resource_view,
                display_mode,
                timeout_minutes=parsed_args.timeout,
                use_color=use_color,
            ).exec()
        except MonitoringError as e:
            uni_print(f"Error monitoring service: {e}", sys.stderr)
            return 1

    def _determine_display_mode(self, requested_mode):
        """Determine and validate the display mode.

        Args:
            requested_mode: User-requested mode ('interactive', 'text-only', or None)

        Returns:
            str: Validated display mode ('interactive' or 'text-only')

        Raises:
            ValueError: If interactive mode is requested without TTY
        """
        # Determine display mode with auto-detection
        if requested_mode is None:
            # Auto-detect: interactive if TTY available, else text-only
            return 'INTERACTIVE' if sys.stdout.isatty() else 'TEXT-ONLY'

        # Validate requested mode
        if requested_mode == 'INTERACTIVE':
            if not sys.stdout.isatty():
                raise ValueError(
                    "Interactive mode requires a TTY (terminal). "
                    "Use --mode TEXT-ONLY for non-interactive environments."
                )
            return 'INTERACTIVE'

        # text-only mode doesn't require TTY
        return requested_mode

    def _should_use_color(self, parsed_globals):
        """Determine if color output should be used based on global settings.

        Args:
            parsed_globals: Global CLI configuration

        Returns:
            bool: True if color should be used
        """
        if parsed_globals.color == 'on':
            return True
        elif parsed_globals.color == 'off':
            return False
        # Default 'auto' behavior - use color if output is a TTY
        return sys.stdout.isatty()


class ECSExpressGatewayServiceWatcher:
    """Monitors ECS Express Gateway Service deployments and resource changes.

    Provides real-time tracking of service deployments with support for different
    monitoring modes (RESOURCE/DEPLOYMENT) and handles resource parsing, status updates,
    and output formatting.

    Args:
        client: ECS client for API calls
        service_arn (str): ARN of the service to monitor
        resource_view (str): Resource view mode - 'RESOURCE' or 'DEPLOYMENT'
        display_mode (str): Display mode - 'INTERACTIVE' or 'TEXT-ONLY'
        timeout_minutes (int): Maximum monitoring time in minutes (default: 30)
    """

    def __init__(
        self,
        client,
        service_arn,
        resource_view,
        display_mode,
        timeout_minutes=30,
        display_strategy=None,
        use_color=True,
        collector=None,
    ):
        self.service_arn = service_arn
        self.display_mode = display_mode
        self.timeout_minutes = timeout_minutes
        self.start_time = time.time()
        self.use_color = use_color
        self.display_strategy = (
            display_strategy or self._create_display_strategy()
        )
        self.collector = collector or ServiceViewCollector(
            client, service_arn, resource_view, use_color
        )

    @staticmethod
    def is_monitoring_available():
        """Check if monitoring is available (requires TTY)."""
        return sys.stdout.isatty()

    def exec(self):
        """Execute monitoring using the appropriate display strategy."""
        self.display_strategy.execute_monitoring(
            collector=self.collector,
            start_time=self.start_time,
            timeout_minutes=self.timeout_minutes,
        )

    def _create_display_strategy(self):
        """Create display strategy based on display mode.

        Returns:
            DisplayStrategy: Appropriate strategy for the selected mode

        Raises:
            ValueError: If display mode is not 'INTERACTIVE' or 'TEXT-ONLY'
        """
        if self.display_mode == 'TEXT-ONLY':
            return TextOnlyDisplayStrategy(use_color=self.use_color)
        elif self.display_mode == 'INTERACTIVE':
            return InteractiveDisplayStrategy(
                display=Display(), use_color=self.use_color
            )
        else:
            raise ValueError(f"Invalid display mode: {self.display_mode}")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Model based auto-completer.

This provides autocompletion based on information found
in the `service-2.json` files.

"""

from collections import namedtuple

from awscli.autocomplete import db

# This module and the awscli.autocomplete.db module are imported
# when a user requests autocompletion.  We should avoid importing
# awscli.clidriver or botocore, which have substantial import
# times.  Autocompleting the command names only needs to load
# the sqlite3 cache file.


CLIArgument = namedtuple(
    'CLIArgument',
    [
        'argname',
        'type_name',
        'command',
        'parent',
        'nargs',
        'positional_arg',
        'required',
    ],
)


class ModelIndex:
    """Retrieve command/param names through querying an index.

    This class provides methods for retrieving valid command
    and parameter names given some context.  It's used by
    the model based autocompleter.

    """

    _COMMAND_NAME_QUERY = """
        SELECT command, full_name FROM command_table
        WHERE parent = :parent
    """

    _ARG_NAME_QUERY = """
        SELECT argname FROM param_table
        WHERE
          parent = :parent AND
          command = :command AND
          positional_arg = :positional_arg
    """

    _ARG_DATA_QUERY = """\
        SELECT argname, type_name, command, parent, nargs, positional_arg,
            required FROM param_table
        WHERE
          parent = :parent AND
          command = :command AND
          argname = :argname
    """

    _GLOBAL_ARG_DATA_QUERY = """
        SELECT argname, type_name, command, parent, nargs, positional_arg,
            required FROM param_table
        WHERE
          parent = :parent
    """

    def __init__(self, db_filename=None):
        self._db_filename = db_filename
        self._db_connection = None

    def _get_db_connection(self):
        if self._db_connection is None:
            self._db_connection = db.DatabaseConnection(self._db_filename)
        return self._db_connection

    def command_names(self, lineage):
        """Return only command names without full_names
            given a lineage.

        :return: A list of available commands
        """
        return [row[0] for row in self.commands_with_full_name(lineage)]

    def commands_with_full_name(self, lineage):
        """Return command names and full_name if it's a service
         given a lineage.

        This ``lineage`` is the same concept used in the
        AWS CLI command classes, except that it explicitly
        uses ``aws`` as the first parent.  This is the list of
        parent commands for a given CLI command.  For example,
        ``aws ec2 <here>`` has a lineage of ``['aws', 'ec2']``.  The
        command ``aws ec2 wait instance-running <here>`` has a lineage
        of ``['aws', 'ec2', 'wait', 'instance-running']``.

        :return: A list of tuples (name, full_name) available commands
        """
        db = self._get_db_connection()
        parent = '.'.join(lineage)
        results = db.execute(self._COMMAND_NAME_QUERY, parent=parent)
        return results.fetchall()

    def arg_names(self, lineage, command_name, positional_arg=False):
        """Return arg names for a given lineage.

        The return values do not have the `--` added, e.g
        we'll return ``region``, not ``-region``.

        If you want the arg names for ``aws ec2 describe-instances``,
        you would provide
        ``lineage=['aws', 'ec2'], command_name='describe-instances'``.

        """
        db = self._get_db_connection()
        parent = '.'.join(lineage)
        results = db.execute(
            self._ARG_NAME_QUERY,
            parent=parent,
            command=command_name,
            positional_arg=positional_arg,
        )
        return [row[0] for row in results]

    def get_argument_data(self, lineage, command_name, arg_name):
        """Return all metadata for a single argument.

        For example, to get the arg data for::

            $ aws s3api list-objects --bucket

        You'd use:

            * ``lineage=['aws', 's3api']``
            * ``command_name='list-objects'``
            * ``arg_name='bucket'``

        :return: A CLIArgument object.

        """
        db = self._get_db_connection()
        parent = '.'.join(lineage)
        results = db.execute(
            self._ARG_DATA_QUERY,
            parent=parent,
            command=command_name,
            argname=arg_name,
        )
        match = results.fetchone()
        if match is not None:
            return CLIArgument(*match)

    def get_global_arg_data(self):
        """Return all metadata for the global args.

        :return: A list of tuples.

        """
        db = self._get_db_connection()
        results = db.execute(self._GLOBAL_ARG_DATA_QUERY, parent='')
        return results.fetchall()



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.exceptions import ProfileNotFound

from awscli.autocomplete.filters import fuzzy_filter
from awscli.autocomplete.main import create_autocompleter
from awscli.autoprompt.prompttoolkit import PromptToolkitPrompter
from awscli.customizations.exceptions import ParamValidationError
from awscli.errorhandler import SilenceParamValidationMsgErrorHandler


class AutoPromptDriver:
    _NO_PROMPT_ARGS = ['help', '--version']
    _CLI_AUTO_PROMPT_OPTION = '--cli-auto-prompt'
    _NO_CLI_AUTO_PROMPT_OPTION = '--no-cli-auto-prompt'

    def __init__(self, driver, completion_source=None, prompter=None):
        self._completion_source = completion_source
        self._prompter = prompter
        self._session = driver.session
        self._driver = driver
        if self._completion_source is None:
            self._completion_source = create_autocompleter(
                driver=self._driver, response_filter=fuzzy_filter
            )

    @property
    def prompter(self):
        if self._prompter is None:
            self._prompter = AutoPrompter(
                self._completion_source, self._driver
            )
        return self._prompter

    def validate_auto_prompt_args_are_mutually_exclusive(self, args):
        no_cli_auto_prompt = self._NO_CLI_AUTO_PROMPT_OPTION in args
        cli_auto_prompt = self._CLI_AUTO_PROMPT_OPTION in args
        if cli_auto_prompt and no_cli_auto_prompt:
            raise ParamValidationError(
                'Both --cli-auto-prompt and --no-cli-auto-prompt cannot be '
                'specified at the same time.'
            )

    def resolve_mode(self, args):
        # Order of precedence to check:
        # - check if any arg rom NO_PROMPT_ARGS in args
        # - check if '--no-cli-auto-prompt' was specified
        # - check if '--cli-auto-prompt' was specified
        # - check configuration chain
        self.validate_auto_prompt_args_are_mutually_exclusive(args)
        if any(arg in args for arg in self._NO_PROMPT_ARGS):
            return 'off'
        if self._NO_CLI_AUTO_PROMPT_OPTION in args:
            return 'off'
        if self._CLI_AUTO_PROMPT_OPTION in args:
            return 'on'
        try:
            config = self._session.get_config_variable('cli_auto_prompt')
            return config.lower()
        except ProfileNotFound:
            return 'off'

    def inject_silence_param_error_msg_handler(self, driver):
        driver.error_handler.inject_handler(
            0, SilenceParamValidationMsgErrorHandler()
        )

    def prompt_for_args(self, args):
        return self.prompter.prompt_for_values(args)


class AutoPrompter:
    """Fills out the arguments list by calling out to the UI prompt backend to
    do the actual prompting of arguments. This makes it easy to swap out
    the UI prompt backend easily if needed.

    """

    def __init__(self, completion_source, driver, prompter=None):
        self._completion_source = completion_source
        self._driver = driver
        if prompter is None:
            prompter = PromptToolkitPrompter(
                self._completion_source, self._driver
            )
        self._prompter = prompter

    def prompt_for_values(self, original_args):
        return self._prompter.prompt_for_args(original_args)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
from functools import partial

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters import has_focus
from prompt_toolkit.formatted_text import HTML, to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_to_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import (
    ConditionalContainer,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.processors import Processor, Transformation
from prompt_toolkit.widgets import (
    Button,
    Dialog,
    Frame,
    HorizontalLine,
    Label,
    TextArea,
)
from prompt_toolkit.widgets.base import Border

from awscli.autoprompt.filters import (
    doc_window_has_focus,
    help_section_visible,
    input_buffer_has_focus,
    is_debug_mode,
    is_history_mode,
    search_input_has_focus,
)


class FormatTextProcessor(Processor):
    """This Processor is used to transform formatted text into a useable
    format inside a ``prompt_toolkit.buffer.Buffer``.

    """

    def apply_transformation(self, text_input):
        # https://python-prompt-toolkit.readthedocs.io/en/master/pages/reference.html#module-prompt_toolkit.formatted_text
        fragments = to_formatted_text(
            HTML(fragment_list_to_text(text_input.fragments))
        )
        return Transformation(fragments)


class TitleLine:
    def __init__(self, title):
        fill = partial(Window, style='class:frame.border')
        self.container = VSplit(
            [
                fill(char=Border.HORIZONTAL),
                fill(width=1, height=1, char='|'),
                Label(
                    title, style='class:frame.label', dont_extend_width=True
                ),
                fill(width=1, height=1, char='|'),
                fill(char=Border.HORIZONTAL),
            ],
            height=1,
        )

    def __pt_container__(self):
        return self.container


class BaseHelpContainer(ConditionalContainer):
    STYLE = '<style fg="darkturquoise">'
    NAME = None
    CONDITION = None
    DIMENSIONS = {}
    FOCUSABLE = False

    def __init__(self):
        window = self.create_window(self.create_buffer())
        super().__init__(window, self.CONDITION)

    @property
    def help_text(self):
        pass

    def create_buffer(self):
        document = Document(text=self.help_text, cursor_position=0)
        help_buffer = Buffer(name=self.NAME, read_only=True)
        help_buffer.set_document(document, bypass_readonly=True)
        return help_buffer

    def create_window(self, help_buffer):
        return Window(
            content=BufferControl(
                buffer=help_buffer,
                input_processors=[FormatTextProcessor()],
                focusable=self.FOCUSABLE,
            ),
            wrap_lines=True,
            **self.DIMENSIONS,
        )


class BaseHelpView(BaseHelpContainer):
    TITLE = None
    DIMENSIONS = {'width': Dimension(max=35, preferred=25)}

    def create_window(self, help_buffer):
        return Frame(
            super(BaseHelpView, self).create_window(help_buffer),
            title=self.TITLE,
        )


class InputHelpView(BaseHelpView):
    TITLE = 'Prompt panel help'
    NAME = 'help_input'
    CONDITION = has_focus('input_buffer')

    @property
    def help_text(self):
        return (
            f'{self.STYLE}[ENTER]</style> Autocomplete Choice/Execute Command\n'
            f'{self.STYLE}[F1]</style> Hide Shortkey Help\n'
            f'{self.STYLE}[F2]</style> Focus on next panel\n'
            f'{self.STYLE}[F3]</style> Hide/Show Docs\n'
            f'{self.STYLE}[F4]</style> One/Multi column prompt\n'
            f'{self.STYLE}[F5]</style> Hide/Show Output\n'
            f'{self.STYLE}[CONTROL+R]</style> On/Off bck-i-search'
        )


class OutputHelpView(BaseHelpView):
    TITLE = 'Output panel help'
    NAME = 'output_input'
    CONDITION = has_focus('output_buffer')

    @property
    def help_text(self):
        return (
            f'{self.STYLE}[/]</style> Search Forward\n'
            f'{self.STYLE}[?]</style> Search Backward\n'
            f'{self.STYLE}[n]</style> Find Next Match\n'
            f'{self.STYLE}[N]</style> Find Previous Match\n'
            f'{self.STYLE}[F1]</style> Hide Shortkey Help\n'
            f'{self.STYLE}[F2]</style> Focus on next panel\n'
            f'{self.STYLE}[F3]</style> Hide/Show Docs\n'
            f'{self.STYLE}[F5]</style> Hide/Show Output\n'
        )


class DocHelpView(BaseHelpView):
    TITLE = 'Doc panel help'
    NAME = 'help_doc'
    CONDITION = doc_window_has_focus | search_input_has_focus

    @property
    def help_text(self):
        return (
            f'{self.STYLE}[/]</style> Search Forward\n'
            f'{self.STYLE}[?]</style> Search Backward\n'
            f'{self.STYLE}[n]</style> Find Next Match\n'
            f'{self.STYLE}[N]</style> Find Previous Match\n'
            f'{self.STYLE}[w]</style> Go Up a Page\n'
            f'{self.STYLE}[z]</style> Go Down a Page\n'
            f'{self.STYLE}[j]</style> Go Up a Line\n'
            f'{self.STYLE}[k]</style> Go Down a Line\n'
            f'{self.STYLE}[g]</style> Go to Top\n'
            f'{self.STYLE}[G]</style> Go to Bottom\n'
            f'{self.STYLE}[F1]</style> Hide Shortkey Help\n'
            f'{self.STYLE}[F2]</style> Focus on next panel\n'
            f'{self.STYLE}[q]</style> Focus on Input\n'
            f'{self.STYLE}[F5]</style> Hide/Show Output'
        )


class BaseToolbarView(BaseHelpContainer):
    DIMENSIONS = {'height': Dimension(max=2)}
    SPACING = '    '


class DocToolbarView(BaseToolbarView):
    NAME = 'toolbar_doc'
    CONDITION = doc_window_has_focus | search_input_has_focus

    @property
    def help_text(self):
        return (
            f'{self.STYLE}[F1]</style> Show Shortkey Help{self.SPACING}'
            f'{self.STYLE}[F2]</style> Focus on next panel{self.SPACING}'
            f'{self.STYLE}[[q]</style> Focus on Input'
        )


class InputToolbarView(BaseToolbarView):
    NAME = 'toolbar_input'
    CONDITION = has_focus('input_buffer')

    @property
    def help_text(self):
        return (
            f'{self.STYLE}[ENTER]</style> Autocomplete '
            f'Choice/Execute Command{self.SPACING}'
            f'{self.STYLE}[F1]</style> Show Shortkey Help{self.SPACING}'
            f'{self.STYLE}[F2]</style> Focus on next panel{self.SPACING}'
            f'{self.STYLE}[F3]</style> Hide/Show Docs{self.SPACING}'
            f'{self.STYLE}[F5]</style> Hide/Show Output'
        )


class OutputToolbarView(BaseToolbarView):
    NAME = 'output_input'
    CONDITION = has_focus('output_buffer')

    @property
    def help_text(self):
        return (
            f'{self.STYLE}[F1]</style> Show Shortkey Help{self.SPACING}'
            f'{self.STYLE}[F2]</style> Focus on next panel{self.SPACING}'
            f'{self.STYLE}[F3]</style> Hide/Show Docs{self.SPACING}'
            f'{self.STYLE}[F5]</style> Hide/Show Output'
        )


class DebugToolbarView(BaseToolbarView):
    NAME = 'toolbar_input'
    CONDITION = is_debug_mode

    @property
    def help_text(self):
        return f'{self.STYLE}[CONTROL+S]</style> Save log to file'


class HistorySignToolbarView(BaseToolbarView):
    NAME = 'history_sign'
    STYLE = '<style fg="ansired" bg="#00ff44">'
    CONDITION = is_history_mode
    DIMENSIONS = {
        'height': Dimension(max=2),
        'width': Dimension(max=13),
    }

    @property
    def help_text(self):
        return f'{self.STYLE}bck-i-search</style>{self.SPACING}'


class ToolbarWidget:
    def __init__(self):
        self.container = HSplit(
            [
                ConditionalContainer(HorizontalLine(), ~help_section_visible),
                VSplit(
                    [
                        HistorySignToolbarView(),
                        ConditionalContainer(
                            VSplit(
                                [
                                    InputToolbarView(),
                                    DocToolbarView(),
                                    OutputToolbarView(),
                                ]
                            ),
                            ~help_section_visible,
                        ),
                    ]
                ),
            ]
        )

    def __pt_container__(self):
        return self.container


class HelpPanelWidget:
    def __init__(self):
        self.container = ConditionalContainer(
            HSplit([DocHelpView(), InputHelpView(), OutputHelpView()]),
            help_section_visible,
        )

    def __pt_container__(self):
        return self.container


class DebugPanelWidget:
    DIMENSIONS = {
        'width': Dimension(max=40, preferred=20),
    }

    def __init__(self):
        self._dialog_prev_focus = None

        self.dialog = self.create_save_file_dialog()

        _kb = KeyBindings()
        _kb.add(Keys.ControlS, filter=is_debug_mode, is_global=True)(
            self._activate_dialog
        )

        self.float_container = FloatContainer(
            Window(
                content=BufferControl(
                    buffer=Buffer(name='debug_buffer', read_only=True)
                ),
                wrap_lines=True,
            ),
            key_bindings=_kb,
            floats=[],
        )
        self.container = ConditionalContainer(
            Frame(
                HSplit(
                    [
                        self.float_container,
                        HorizontalLine(),
                        DebugToolbarView(),
                    ]
                ),
                **self.DIMENSIONS,
                title='Debug panel',
            ),
            filter=is_debug_mode,
        )

    def _activate_dialog(self, event):
        layout = event.app.layout
        self._dialog_prev_focus = event.app.current_buffer
        self.float_container.floats.append(Float(content=self.dialog))
        layout.focus(self.dialog)

    def create_save_file_dialog(self):
        default_filename = 'prompt_debug.log'

        def cancel_handler():
            app = get_app()
            textfield.text = default_filename
            app.layout.focus(textfield)
            self.float_container.floats.pop()
            app.layout.focus(self._dialog_prev_focus)

        def ok_handler(*args, **kwargs):
            app = get_app()
            buffer = app.layout.get_buffer_by_name('debug_buffer')
            with open(os.path.expanduser(textfield.text), 'w') as f:
                f.write(buffer.document.text)
            textfield.text = default_filename
            app.layout.focus(textfield)
            self.float_container.floats.pop()
            app.layout.focus(self._dialog_prev_focus)

        ok_button = Button(text='Save', handler=ok_handler)
        cancel_button = Button(text='Cancel', handler=cancel_handler)

        textfield = TextArea(text=default_filename, multiline=False)

        dialog = Dialog(
            title='Save logs to file',
            body=HSplit(
                [
                    Label(text='Log file name', dont_extend_height=True),
                    textfield,
                ],
                padding=Dimension(preferred=1, max=1),
            ),
            buttons=[ok_button, cancel_button],
            with_background=True,
        )
        # add keybinding to save file on press Enter in textfield
        dialog.container.body.container.content.key_bindings.add(
            Keys.Enter, filter=has_focus(textfield)
        )(ok_handler)

        return dialog

    def __pt_container__(self):
        return self.container



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.layout.containers import Dimension, Window
from prompt_toolkit.layout.controls import BufferControl


def get_ui_control_by_buffer_name(layout, buffer_name):
    """Retrieves any UI control based on it's buffer name

    Prompt toolkit has a variety of methods on its layout to retrieve the
    a buffer by name but most of them require the control to be a
    BufferControl object. This function allows for better duck typing in that
    it includes UI control objects that have buffer properties.
    """
    for control in layout.find_all_controls():
        if hasattr(control, 'buffer') and control.buffer.name == buffer_name:
            return control
    raise ValueError(
        f"Couldn't find buffer in the current layout: {buffer_name}"
    )


def move_to_previous_prompt(app):
    """Moves to traverser's previous prompt in the layout"""
    previous_prompt = app.traverser.previous_prompt()
    show_details_if_visible_by_default(app, previous_prompt)
    refresh_details_view(app, previous_prompt)
    previous_control = get_ui_control_by_buffer_name(
        app.layout, previous_prompt
    )
    app.layout.focus(previous_control)


def show_details_if_visible_by_default(app, prompt):
    app.details_visible = app.traverser.is_prompt_details_visible_by_default(
        prompt
    )


def refresh_details_view(app, prompt):
    control = get_ui_control_by_buffer_name(app.layout, prompt)
    if callable(getattr(control, 'on_toggle', None)):
        control.on_toggle(control.buffer.text)


class Spacer:
    """Fills empty space in a layout

    It is helpful for extending a particular container with a visual
    element such as expanding tab column in the wizard app with the
    color gray.
    """

    def __init__(self):
        self.container = self._get_container()

    def _get_container(self):
        buffer = Buffer(document=Document(''), read_only=True)
        return Window(content=BufferControl(buffer=buffer, focusable=False))

    def __pt_container__(self):
        return self.container


class FullyExtendedWidthWindow(Window):
    """Window that fully extends its available width"""

    def preferred_width(self, max_available_width):
        return Dimension(preferred=max_available_width)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class ParamValidationError(Exception):
    """CLI parameter validation failed. Indicates RC 252.

    This exception indicates that the command was either invalid or failed to
    pass a client side validation on the command syntax or parameters provided.
    """


class ConfigurationError(Exception):
    """CLI configuration is an invalid state. Indicates RC 253.

    This exception indicates that the command run may be syntactically correct
    but the CLI's environment or configuration is incorrect, incomplete, etc.
    """



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
WORD_BOUNDARY = ''


class ParsedResult:
    def __init__(
        self,
        current_command=None,
        current_param=None,
        global_params=None,
        parsed_params=None,
        lineage=None,
        current_fragment=None,
        unparsed_items=None,
    ):
        """

        :param current_command: The name of the leaf command; the most
            specific subcommand that was parsed.
        :param current_param: The name of the current parameter; this is the
            last known parameter we've parsed.  This value is set when we
            have parsed a known parameter and are parsing values associated
            with this parameter, and will be set to None once we're no longer
            parsing parameter values for the param.
        :param global_params: A dictionary of global CLI params.
        :param parsed_params: A dictionary of parameters that apply to
            the current command (non global params).
        :param lineage: A list of the lineage, with the ``aws`` portion
            included as the first element if necessary.
        :param current_fragment: The last fragment of a word that was not
            parsed.  This can happen if the user types only part of a command.
            The ``current_fragment`` is only populated if the cursor is on
            the current word.  If there is a space between the last word
            and the cursor it is not considered a fragment.  If this value
            is non ``None``, it typically indicates the value that needs to
            be auto-completed.
        :param unparsed_items: A list of items that we were not able to
            parse.  This can happen if the user types a command that's not
            in the index (e.g ``aws foo bar``).

        """
        # Example:
        #  aws --debug ec2 describe-instances --instance-ids i-123 i-124 \
        #    --region us-west-2
        #
        # This would parse to:
        # ParsedResult(
        #  current_command='describe-instances',
        #  global_params={'debug': None, 'region': 'us-west-2'},
        #  parsed_params={'instance-ids': ['i-123', 'i-124']},
        #  lineage=['aws', 'ec2'],
        #  current_fragment=None
        # )
        #
        # An example of last fragment is:
        #
        # "aws ec2 describe-"
        #
        # This would parse to:
        #
        # ParsedResult(current_command='ec2', lineage=['aws'],
        #              current_fragment='describe-')
        self.current_command = current_command
        self.current_param = current_param
        if global_params is None:
            global_params = {}
        self.global_params = global_params
        if parsed_params is None:
            parsed_params = {}
        self.parsed_params = parsed_params
        if lineage is None:
            lineage = []
        self.lineage = lineage
        # current_fragment is used to indicate that the value
        # is not a known subcommand/option.  It will only
        # ever apply to the last word in the command line.
        self.current_fragment = current_fragment
        if unparsed_items is None:
            unparsed_items = []
        self.unparsed_items = unparsed_items

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__


class ParseState:
    def __init__(self):
        self._current_command = None
        self.current_param = None
        self._lineage = []

    @property
    def current_command(self):
        return self._current_command

    @current_command.setter
    def current_command(self, value):
        if self._current_command is not None:
            self._lineage.append(self._current_command)
        self._current_command = value

    @property
    def lineage(self):
        return self._lineage

    @property
    def full_lineage(self):
        if self.current_command is None:
            return self._lineage
        return self._lineage + [self.current_command]


class CLIParser:
    """Parses AWS CLI command.

    This is different from the parser used when actually invoking CLI commands
    in that it can handle incomplete commands.  It will partially parse
    everything it understands, but incomplete commands will not generate
    errors.  This class is intended for use with auto completion and is
    not a general purpose AWS CLI parser.

    """

    def __init__(self, index, return_first_command_match=False):
        self._index = index
        self._return_first_command_match = return_first_command_match

    def parse(self, command_line, location=None):
        """Parses as much of the command line input as possible.

        :param command_line: The command line as a string.
        :param location: The offset (0 based) of the current location
            of the cursor.

        """
        # NOTE: This parser doesn't support using `=` as a separator
        # between option and value, e.g `--foo=bar`.  This is something
        # we should look into adding.
        parsed = ParsedResult()
        state, remaining_parts = self._split_to_parts(command_line, location)
        global_args = self._index.arg_names(lineage=[], command_name='aws')
        current_args = []
        current = state.current_command
        while remaining_parts:
            current = remaining_parts.pop(0)
            if current.startswith('--'):
                self._handle_option(
                    current,
                    remaining_parts,
                    current_args,
                    global_args,
                    parsed,
                    state,
                )
            else:
                current_args = self._handle_positional(
                    current, state, remaining_parts, parsed
                )
        parsed.current_command = state.current_command
        parsed.current_param = state.current_param
        parsed.lineage = state.lineage
        return parsed

    def _consume_value(
        self, remaining_parts, option_name, lineage, current_command, state
    ):
        # We have a special case where a user is trying to complete
        # a value for an option, which is the last fragment of the command,
        # e.g. 'aws ec2 describe-instances --instance-ids '
        # Note the space at the end.  In this case we don't have a value
        # to consume so we special case this and short circuit.
        arg_data = self._index.get_argument_data(
            lineage=lineage,
            command_name=current_command,
            arg_name=option_name,
        )
        if arg_data.type_name == 'boolean':
            state.current_param = None
            return True
        elif remaining_parts == [WORD_BOUNDARY]:
            return ''
        elif len(remaining_parts) <= 1:
            return None
        nargs = arg_data.nargs
        # We're making an assumption about nargs in order to simplify
        # its handling.  Handling nargs='*' and nargs='+' normally takes into
        # account specific 'stop words', in our case valid subcommands.  This
        # means we technically support edge cases such as
        # "aws ec2 --filters describe-instances", which will parse to
        # `['aws', 'ec2', 'describe-instances']`, {'filters': []}.  Our
        # assumption is non-global options typically come after the associated
        # subcommand, e.g. "aws ec2 describe-instances <params here>".  We
        # don't have to worry about global params because none of them use
        # nargs '*' or '+'.
        if nargs is None:
            # The default behavior is to consume a single arg.
            result = remaining_parts.pop(0)
            state.current_param = None
            return result
        elif nargs == '?':
            if remaining_parts and not remaining_parts[0].startswith('--'):
                result = remaining_parts.pop(0)
                state.current_param = None
                return result
        elif nargs in '+*':
            # Technically nargs='+' requires one or more args, but
            # we don't validate this.  This will just result in
            # an empty list being returned.  This is acceptable
            # for auto-completion purposes.
            value = []
            while len(remaining_parts) > 0 and not remaining_parts == [
                WORD_BOUNDARY
            ]:
                if remaining_parts[0].startswith('--'):
                    state.current_param = None
                    break
                if len(remaining_parts) == 1:
                    break
                value.append(remaining_parts.pop(0))
            return value

    def _split_to_parts(self, command_line, location):
        state = ParseState()
        if location is not None:
            # The original auto completer had this logic.
            # We could in theory still parse the entire command
            # line and only use the location value for determining
            # what to auto complete, but we're keeping the same logic
            # as the original for now.
            command_line = command_line[:location]
        parts = command_line.split()
        if command_line and command_line[-1].isspace():
            # If the command line ends with whitespace then we insert
            # an empty element in the parts list to ensure that the
            # last chunk isn't marked as the last fragment.  This is
            # because we can auto-complete a command such as:
            # "aws ec2 stop-<TAB>" but we can't auto-complete
            # a command like: "aws ec2 stop- <TAB>"
            parts.append(WORD_BOUNDARY)
        if parts:
            # If parts are returned, we are assuming that the first argument
            # will always be the executable name. However the executable name
            # can be anything (e.g. aws2), so we normalize the first
            # current_command to be aws as the autocomplete index expects the
            # first command to be aws.
            parts.pop(0)
            state.current_command = 'aws'
        return state, parts

    def _handle_option(
        self,
        current,
        remaining_parts,
        current_args,
        global_args,
        parsed,
        state,
    ):
        if current_args is None:
            # If there are no arguments found for this current scope,
            # it usually indicates we've encounted a command we don't know.
            # In this case we don't try to handle this option.
            parsed.unparsed_items.append(current)
            return
        # This is a command line option, remove the `--` portion so we
        # just have the option name.
        option_name = current[2:]
        if option_name in global_args:
            state.current_param = option_name
            value = self._consume_value(
                remaining_parts,
                option_name,
                lineage=[],
                state=state,
                current_command='aws',
            )
            parsed.global_params[option_name] = value
        elif option_name in current_args:
            state.current_param = option_name
            value = self._consume_value(
                remaining_parts,
                option_name,
                state.lineage,
                state.current_command,
                state=state,
            )
            parsed.parsed_params[option_name] = value
        elif self._is_last_word(remaining_parts, current):
            # If the option wasn't handled, there's two cases.
            # Either it's the last word (without a space) so it's
            # something we should auto-complete.  In that case
            # we denote it as the last fragment.
            parsed.current_fragment = current
        else:
            # It's not a known option in our index and it's not
            # a partially completed words.  This goes into the
            # unparsed_items list.
            parsed.unparsed_items.append(current)

    def _is_last_word(self, remaining_parts, current):
        return not remaining_parts and current

    def _is_part_of_command(self, current, command_names):
        return any(
            command.startswith(current) and command != current
            for command in command_names
        )

    def _is_command_name(self, current, remaining_parts, command_names):
        # If _return_first_command_match is True
        #
        # We just check if the 'current' is in 'command_names'
        # and don't care if there any other commands that could have 'current'
        # as a part
        #
        # If _return_first_command_match is False
        #
        # To decide if 'current' part is a command or not we consider such cases
        # - if 'current' is in 'command_names' and we already moved forward (at
        # least entered ' ' after) then it's a command
        # - if 'current' is in 'command_names' and we have no other commands
        # started with this prefix then it's a command
        #
        # but if 'current' is in 'command_names' but we have other commands
        # start with the same prefix, for example 's3' and 's3api' we don't
        # consider 'current' as a complete command
        is_command_name = current in command_names
        if self._return_first_command_match:
            return is_command_name
        is_part_of_command = self._is_part_of_command(current, command_names)
        return is_command_name and (remaining_parts or not is_part_of_command)

    def _handle_positional(self, current, state, remaining_parts, parsed):
        # This is can either be a subcommand or a positional argument
        #
        # First we can check if this is a valid subcommand given our lineage.
        # We're one off here, we need to compute a new *potential*
        # lineage.
        command_names = self._index.command_names(state.full_lineage)
        positional_argname = None
        if self._is_command_name(current, remaining_parts, command_names):
            state.current_command = current
            # We also need to get the next set of command line options.
            current_args = self._index.arg_names(
                lineage=state.lineage, command_name=state.current_command
            )
            return current_args
        if not command_names:
            # If there are no more command names check. See if the command
            # has a positional argument. This will require an additional
            # select on the argument index.
            positional_argname = self._get_positional_argname(state)
        if (
            positional_argname
            and positional_argname not in parsed.parsed_params
        ):
            # Parse the current string to be a positional argument
            # if the command has the a positional arg and the positional arg
            # has not already been parsed.
            if not remaining_parts:
                # We are currently parsing the positional argument
                parsed.current_fragment = current
                state.current_param = positional_argname
            elif current:
                # At this point, the positional argument is fully
                # written out so it should be saved to the parsed
                # params
                parsed.parsed_params[positional_argname] = current
                state.current_param = None
            return self._index.arg_names(
                lineage=state.lineage, command_name=state.current_command
            )
        else:
            if not remaining_parts:
                # If this is the last chunk of the command line but
                # it's not a subcommand then we'll mark it as the last
                # fragment.  This is likely a partially entered
                # command, e.g 'aws ec2 run-instan'
                parsed.current_fragment = current
            elif current:
                outfile_arg = self._index.get_argument_data(
                    lineage=state.lineage,
                    command_name=state.current_command,
                    arg_name='outfile',
                )
                if outfile_arg and 'outfile' not in parsed.parsed_params:
                    # if "outfile" is in command args we need to handle it in a
                    # special way because "outfile" is a special option which
                    # is marked as option but actually a positional argument
                    # that can be at any place in the command line
                    parsed.parsed_params['outfile'] = current
                    state.current_param = None
                    return self._index.arg_names(
                        lineage=state.lineage,
                        command_name=state.current_command,
                    )
                else:
                    # Otherwise this is some command we don't know about
                    # so we add it to the list of unparsed_items.
                    parsed.unparsed_items.append(current)
            return None

    def _get_positional_argname(self, state):
        positional_args = self._index.arg_names(
            lineage=state.lineage,
            command_name=state.current_command,
            positional_arg=True,
        )
        if positional_args:
            # We are assuming there is only ever one positional
            # argument for a command.
            return positional_args[0]
        return None
